"""Decorator for creating a run tree from functions."""

from __future__ import annotations

import asyncio
import contextlib
import contextvars
import datetime
import functools
import inspect
import logging
import threading
import warnings
import weakref
from collections.abc import (
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Generator,
    Iterator,
    Mapping,
    Sequence,
)
from contextvars import copy_context
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    Protocol,
    TypedDict,
    TypeVar,
    Union,
    cast,
    get_type_hints,
    overload,
    runtime_checkable,
)

from typing_extensions import ParamSpec, TypeGuard, get_args, get_origin

import langsmith._internal._context as _context
from langsmith import client as ls_client
from langsmith import run_trees, schemas, utils
from langsmith._internal import _aiter as aitertools
from langsmith._runtime_overrides import (
    _aio_to_thread_override_active as _runtime_override_active,
)
from langsmith.env import _runtime_env
from langsmith.run_trees import WriteReplica

if TYPE_CHECKING:
    from types import TracebackType

    from langchain_core.runnables import Runnable

LOGGER = logging.getLogger(__name__)
_CONTEXT_KEYS: dict[str, contextvars.ContextVar] = {
    "parent_ref": _context._PARENT_RUN_TREE_REF,
    "project_name": _context._PROJECT_NAME,
    "tags": _context._TAGS,
    "metadata": _context._METADATA,
    "enabled": _context._TRACING_ENABLED,
    "client": _context._CLIENT,
    "replicas": run_trees._REPLICAS,
    "distributed_parent_id": run_trees._DISTRIBUTED_PARENT_ID,
}

_EXCLUDED_FRAME_FNAME = "langsmith/run_helpers.py"

_OTEL_AVAILABLE: Optional[bool] = None


def get_current_run_tree() -> Optional[run_trees.RunTree]:
    """Get the current run tree.

    Uses a weakref-based lookup to avoid memory leaks from captured contexts.
    The RunTree may return None if it has been garbage collected.
    """
    return _context.get_current_run_tree()


@contextlib.contextmanager
def set_tracing_parent(
    run_tree: run_trees.RunTree,
) -> Generator[None, None, None]:
    """Set a RunTree as the active tracing parent within this block.

    Unlike `tracing_context`, this only sets the parent run tree and nothing
    else, making it safe to use in isolated threads where you want precise
    control over which run acts as the parent without inheriting or overwriting
    other context variables.

    Args:
        run_tree: The RunTree to use as the active parent.
    """
    token = _context._PARENT_RUN_TREE_REF.set(weakref.ref(run_tree))
    try:
        yield
    finally:
        _context._PARENT_RUN_TREE_REF.reset(token)


def set_run_metadata(**metadata: Any) -> None:
    """Update metadata on the current run tree."""
    run_tree = get_current_run_tree()
    if run_tree is None:
        LOGGER.warning(
            "No active run tree found. Call `set_run_metadata` inside a traced run."
        )
    else:
        run_tree.metadata.update(metadata)
    return


def get_tracing_context(
    context: Optional[contextvars.Context] = None,
) -> dict[str, Any]:
    """Get the current tracing context."""
    if context is None:
        parent = _context.get_current_run_tree()
        return {
            "parent": parent,
            "project_name": _context._PROJECT_NAME.get(),
            "tags": _context._TAGS.get(),
            "metadata": _context._METADATA.get(),
            "enabled": _context._TRACING_ENABLED.get(),
            "client": _context._CLIENT.get(),
            "replicas": run_trees._REPLICAS.get(),
            "distributed_parent_id": run_trees._DISTRIBUTED_PARENT_ID.get(),
        }
    # When reading from a copied context, dereference the weakref
    result = {k: context.get(v) for k, v in _CONTEXT_KEYS.items()}
    parent_ref = result.pop("parent_ref", None)
    result["parent"] = parent_ref() if parent_ref is not None else None
    return result


@contextlib.contextmanager
def tracing_context(
    *,
    project_name: Optional[str] = None,
    tags: Optional[list[str]] = None,
    metadata: Optional[dict[str, Any]] = None,
    parent: Optional[Union[run_trees.RunTree, Mapping, str, Literal[False]]] = None,
    enabled: Optional[Union[bool, Literal["local"]]] = None,
    client: Optional[ls_client.Client] = None,
    replicas: Optional[Sequence[WriteReplica]] = None,
    distributed_parent_id: Optional[str] = None,
    **kwargs: Any,
) -> Generator[None, None, None]:
    """Set the tracing context for a block of code.

    Args:
        project_name: The name of the project to log the run to.
        tags: The tags to add to the run.
        metadata: The metadata to add to the run.
        parent: The parent run to use for the context.

            Can be a Run/`RunTree` object, request headers (for distributed tracing),
            or the dotted order string.
        client: The client to use for logging the run to LangSmith.
        enabled: Whether tracing is enabled.

            Defaults to `None`, meaning it will use the current context value or environment variables.
        replicas: A sequence of `WriteReplica` dictionaries to send runs to.

            Example: `[{"api_url": "https://api.example.com", "auth": {"api_key": "key"}, "project_name": "proj"}]`
            or `[{"project_name": "my_experiment", "updates": {"reference_example_id": None}}]`
        distributed_parent_id: The distributed parent ID for distributed tracing. Defaults to None.
    """
    if kwargs:
        # warn
        warnings.warn(
            f"Unrecognized keyword arguments: {kwargs}.",
            DeprecationWarning,
        )
    current_context = get_tracing_context()
    parent_run = (
        _get_parent_run({"parent": parent or kwargs.get("parent_run")})
        if parent is not False
        else None
    )
    distributed_parent_id_to_use = distributed_parent_id
    if distributed_parent_id_to_use is None and parent_run is not None:
        # TODO(angus): decide if we want to merge tags and metadata
        tags = sorted(set(tags or []) | set(parent_run.tags or []))
        metadata = {**parent_run.metadata, **(metadata or {})}
        distributed_parent_id_to_use = parent_run.id  # type: ignore[assignment]
    enabled = enabled if enabled is not None else current_context.get("enabled")
    _set_tracing_context(
        {
            "parent": parent_run,
            "project_name": project_name,
            "tags": tags,
            "metadata": metadata,
            "enabled": enabled,
            "client": client,
            "replicas": replicas,
            "distributed_parent_id": distributed_parent_id_to_use,
        }
    )
    try:
        yield
    finally:
        _set_tracing_context(current_context)


# Alias for backwards compatibility
get_run_tree_context = get_current_run_tree


def is_traceable_function(func: Any) -> TypeGuard[SupportsLangsmithExtra[P, R]]:
    """Check if a function is `@traceable` decorated."""
    return (
        _is_traceable_function(func)
        or (isinstance(func, functools.partial) and _is_traceable_function(func.func))
        or (hasattr(func, "__call__") and _is_traceable_function(func.__call__))
    )


def ensure_traceable(
    func: Callable[P, R],
    *,
    name: Optional[str] = None,
    metadata: Optional[Mapping[str, Any]] = None,
    tags: Optional[list[str]] = None,
    client: Optional[ls_client.Client] = None,
    reduce_fn: Optional[Callable[[Sequence], Union[dict, str]]] = None,
    project_name: Optional[str] = None,
    process_inputs: Optional[Callable[[dict], dict]] = None,
    process_outputs: Optional[Callable[..., dict]] = None,
    process_chunk: Optional[Callable] = None,
) -> SupportsLangsmithExtra[P, R]:
    """Ensure that a function is traceable."""
    if is_traceable_function(func):
        return func
    return traceable(
        name=name,
        metadata=metadata,
        tags=tags,
        client=client,
        reduce_fn=reduce_fn,
        project_name=project_name,
        process_inputs=process_inputs,
        process_outputs=process_outputs,
        process_chunk=process_chunk,
    )(func)


def is_async(func: Callable) -> bool:
    """Inspect function or wrapped function to see if it is async."""
    return inspect.iscoroutinefunction(func) or (
        hasattr(func, "__wrapped__") and inspect.iscoroutinefunction(func.__wrapped__)
    )


class LangSmithExtra(TypedDict, total=False):
    """Any additional info to be injected into the run dynamically."""

    name: Optional[str]
    """Optional name for the run."""
    reference_example_id: Optional[ls_client.ID_TYPE]
    """Optional ID of a reference example."""
    run_extra: Optional[dict]
    """Optional additional run information."""
    parent: Optional[Union[run_trees.RunTree, str, Mapping]]
    """Optional parent run, can be a RunTree, string, or mapping."""
    run_tree: Optional[run_trees.RunTree]  # TODO: Deprecate
    """Optional run tree (deprecated)."""
    project_name: Optional[str]
    """Optional name of the project."""
    metadata: Optional[dict[str, Any]]
    """Optional metadata for the run."""
    tags: Optional[list[str]]
    """Optional list of tags for the run."""
    run_id: Optional[ls_client.ID_TYPE]
    """Optional ID for the run."""
    client: Optional[ls_client.Client]
    """Optional LangSmith client."""
    # Optional callback function to be called if the run succeeds and before it is sent.
    _on_success: Optional[Callable[[run_trees.RunTree], None]]
    on_end: Optional[Callable[[run_trees.RunTree], Any]]
    """Optional callback function to be called after the run ends and is sent."""


R = TypeVar("R", covariant=True)
P = ParamSpec("P")


@runtime_checkable
class SupportsLangsmithExtra(Protocol, Generic[P, R]):
    """Implementations of this Protocol accept an optional langsmith_extra parameter."""

    def __call__(  # type: ignore[valid-type]
        self,
        *args: P.args,
        langsmith_extra: Optional[LangSmithExtra] = None,
        **kwargs: P.kwargs,
    ) -> R:
        """Call the instance when it is called as a function.

        Args:
            *args: Variable length argument list.
            langsmith_extra: Optional dictionary containing additional
                parameters specific to Langsmith.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            R: The return value of the method.

        """
        ...


def _extract_usage(
    *,
    run_tree: run_trees.RunTree,
    outputs: Optional[dict] = None,
    **kwargs: Any,
) -> Optional[schemas.ExtractedUsageMetadata]:
    from_metadata = (run_tree.metadata or {}).get("usage_metadata")
    return (outputs or {}).get("usage_metadata") or from_metadata


@overload
def traceable(
    func: Callable[P, R],
) -> SupportsLangsmithExtra[P, R]: ...


@overload
def traceable(
    run_type: ls_client.RUN_TYPE_T = "chain",
    *,
    name: Optional[str] = None,
    metadata: Optional[Mapping[str, Any]] = None,
    tags: Optional[list[str]] = None,
    client: Optional[ls_client.Client] = None,
    reduce_fn: Optional[Callable[[Sequence], Union[dict, str]]] = None,
    project_name: Optional[str] = None,
    process_inputs: Optional[Callable[[dict], dict]] = None,
    process_outputs: Optional[Callable[..., dict]] = None,
    process_chunk: Optional[Callable] = None,
    _invocation_params_fn: Optional[Callable[[dict], dict]] = None,
    dangerously_allow_filesystem: bool = False,
    enabled: Optional[bool] = None,
    exceptions_to_handle: Optional[tuple[type[BaseException], ...]] = None,
) -> Callable[[Callable[P, R]], SupportsLangsmithExtra[P, R]]: ...


def traceable(
    *args: Any,
    **kwargs: Any,
) -> Union[Callable, Callable[[Callable], Callable]]:
    """Trace a function with langsmith.

    Args:
        run_type: The type of run (span) to create.

            Examples: `llm`, `chain`, `tool`, `prompt`, `retriever`, etc.

            Defaults to "chain".
        name: The name of the run. Defaults to the function name.
        metadata: The metadata to add to the run. Defaults to `None`.
        tags: The tags to add to the run. Defaults to `None`.
        client: The client to use for logging the run to LangSmith. Defaults to
            `None`, which will use the default client.
        reduce_fn: A function to reduce the output of the function if the function
            returns a generator.

            Defaults to `None`, which means the values will be logged as a list.

            !!! note

                If the iterator is never exhausted (e.g. the function returns an
                infinite generator), this will never be called, and the run itself will
                be stuck in a pending state.
        project_name: The name of the project to log the run to.

            Defaults to `None`, which will use the default project.
        process_inputs: Custom serialization / processing function for inputs.

            Defaults to `None`.
        process_outputs: Custom serialization / processing function for outputs.

            Defaults to `None`.
        dangerously_allow_filesystem: Whether to allow filesystem access for attachments.

            Defaults to `False`.

            Traces that reference local filepaths will be uploaded to LangSmith.
            In general, network-hosted applications should not be using this because
            referenced files are usually on the user's machine, not the host machine.
        enabled: Whether tracing is enabled for this function.

            Defaults to `None`, which will use the default value from the current context.
        exceptions_to_handle: Exception types to ignore when logging errors.

            If an exception of one of these types is raised, the run will still be recorded
            but the error field will be `None` instead of containing the full traceback.

            Defaults to `None`.

    Returns:
        The decorated function.

    !!! note

        Requires that `LANGSMITH_TRACING_V2` be set to 'true' in the environment.

    Examples:
        !!! example "Basic usage"

            ```python
            @traceable
            def my_function(x: float, y: float) -> float:
                return x + y


            my_function(5, 6)


            @traceable
            async def my_async_function(query_params: dict) -> dict:
                async with httpx.AsyncClient() as http_client:
                    response = await http_client.get(
                        "https://api.example.com/data",
                        params=query_params,
                    )
                    return response.json()


            asyncio.run(my_async_function({"param": "value"}))
            ```

        !!! example "Streaming data with a generator"

            ```python
            @traceable
            def my_generator(n: int) -> Iterable:
                for i in range(n):
                    yield i


            for item in my_generator(5):
                print(item)
            ```

        !!! example "Async streaming data"

            ```python
            @traceable
            async def my_async_generator(query_params: dict) -> Iterable:
                async with httpx.AsyncClient() as http_client:
                    response = await http_client.get(
                        "https://api.example.com/data",
                        params=query_params,
                    )
                    for item in response.json():
                        yield item


            async def async_code():
                async for item in my_async_generator({"param": "value"}):
                    print(item)


            asyncio.run(async_code())
            ```

        !!! example "Specifying a run type and name"

            ```python
            @traceable(name="CustomName", run_type="tool")
            def another_function(a: float, b: float) -> float:
                return a * b


            another_function(5, 6)
            ```

        !!! example "Logging with custom metadata and tags"

            ```python
            @traceable(
                metadata={"version": "1.0", "author": "John Doe"}, tags=["beta", "test"]
            )
            def tagged_function(x):
                return x**2


            tagged_function(5)
            ```

        !!! example "Specifying a custom client and project name"

            ```python
            custom_client = Client(api_key="your_api_key")


            @traceable(client=custom_client, project_name="My Special Project")
            def project_specific_function(data):
                return data


            project_specific_function({"data": "to process"})
            ```

        !!! example "Manually passing `langsmith_extra`"

            ```python
            @traceable
            def manual_extra_function(x):
                return x**2


            manual_extra_function(5, langsmith_extra={"metadata": {"version": "1.0"}})
            ```

        !!! example "Handling specific exceptions"

            ```python
            @traceable(exceptions_to_handle=(ValueError, TypeError))
            def function_with_handled_exceptions(x):
                if x < 0:
                    raise ValueError("Negative value")  # Won't send error in the trace
                return x**2


            function_with_handled_exceptions(-5)
            ```
    """
    run_type = cast(
        ls_client.RUN_TYPE_T,
        (
            args[0]
            if args and isinstance(args[0], str)
            else (kwargs.pop("run_type", None) or "chain")
        ),
    )
    if run_type not in _VALID_RUN_TYPES:
        warnings.warn(
            f"Unrecognized run_type: {run_type}. Must be one of: {_VALID_RUN_TYPES}."
            f" Did you mean @traceable(name='{run_type}')?"
        )
    if len(args) > 1:
        warnings.warn(
            "The `traceable()` decorator only accepts one positional argument, "
            "which should be the run_type. All other arguments should be passed "
            "as keyword arguments."
        )
    if "extra" in kwargs:
        warnings.warn(
            "The `extra` keyword argument is deprecated. Please use `metadata` "
            "instead.",
            DeprecationWarning,
        )
    reduce_fn = kwargs.pop("reduce_fn", None)
    enabled = kwargs.pop("enabled", None)
    container_input = _ContainerInput(
        # TODO: Deprecate raw extra
        extra_outer=kwargs.pop("extra", None),
        name=kwargs.pop("name", None),
        metadata=kwargs.pop("metadata", None),
        tags=kwargs.pop("tags", None),
        client=kwargs.pop("client", None),
        project_name=kwargs.pop("project_name", None),
        run_type=run_type,
        process_inputs=kwargs.pop("process_inputs", None),
        process_chunk=kwargs.pop("process_chunk", None),
        invocation_params_fn=kwargs.pop("_invocation_params_fn", None),
        dangerously_allow_filesystem=kwargs.pop("dangerously_allow_filesystem", False),
        enabled=enabled,
        exceptions_to_handle=kwargs.pop("exceptions_to_handle", None),
    )
    outputs_processor = kwargs.pop("process_outputs", None)
    _on_run_end = functools.partial(
        _handle_container_end,
        outputs_processor=outputs_processor,
    )

    if kwargs:
        warnings.warn(
            f"The following keyword arguments are not recognized and will be ignored: "
            f"{sorted(kwargs.keys())}.",
            DeprecationWarning,
        )

    def decorator(func: Callable):
        func_sig = inspect.signature(func)
        func_accepts_parent_run = func_sig.parameters.get("run_tree", None) is not None
        func_accepts_config = func_sig.parameters.get("config", None) is not None

        @functools.wraps(func)
        async def async_wrapper(
            *args: Any,
            langsmith_extra: Optional[LangSmithExtra] = None,
            **kwargs: Any,
        ) -> Any:
            """Async version of wrapper function."""
            if not func_accepts_config:
                kwargs.pop("config", None)
            run_container = await aitertools.aio_to_thread(
                copy_context(),
                _setup_run,
                func,
                container_input=container_input,
                langsmith_extra=langsmith_extra,
                args=args,
                kwargs=kwargs,
            )

            try:
                accepts_context = aitertools.asyncio_accepts_context()
                if func_accepts_parent_run:
                    kwargs["run_tree"] = run_container["new_run"]

                otel_context_manager = _maybe_create_otel_context(
                    run_container["new_run"]
                )
                use_ctx_task = accepts_context and not _runtime_override_active()
                if otel_context_manager:

                    async def run_with_otel_context():
                        with otel_context_manager:
                            return await func(*args, **kwargs)

                    if use_ctx_task:
                        function_result = await asyncio.create_task(  # type: ignore[call-arg]
                            run_with_otel_context(), context=run_container["context"]
                        )
                    else:
                        with tracing_context(
                            **get_tracing_context(run_container["context"])
                        ):
                            function_result = await run_with_otel_context()
                else:
                    fr_coro = func(*args, **kwargs)
                    if use_ctx_task:
                        function_result = await asyncio.create_task(  # type: ignore[call-arg]
                            fr_coro, context=run_container["context"]
                        )
                    else:
                        with tracing_context(
                            **get_tracing_context(run_container["context"])
                        ):
                            function_result = await fr_coro
            except BaseException as e:
                # shield from cancellation, given we're catching all exceptions
                _cleanup_traceback(e)
                await asyncio.shield(
                    aitertools.aio_to_thread(
                        copy_context(), _on_run_end, run_container, error=e
                    )
                )
                raise
            await aitertools.aio_to_thread(
                copy_context(), _on_run_end, run_container, outputs=function_result
            )
            return function_result

        @functools.wraps(func)
        async def async_generator_wrapper(
            *args: Any, langsmith_extra: Optional[LangSmithExtra] = None, **kwargs: Any
        ) -> AsyncGenerator:
            if not func_accepts_config:
                kwargs.pop("config", None)
            run_container = await aitertools.aio_to_thread(
                copy_context(),
                _setup_run,
                func,
                container_input=container_input,
                langsmith_extra=langsmith_extra,
                args=args,
                kwargs=kwargs,
            )
            results: list[Any] = []
            try:
                if func_accepts_parent_run:
                    kwargs["run_tree"] = run_container["new_run"]
                    # TODO: Nesting is ambiguous if a nested traceable function is only
                    # called mid-generation. Need to explicitly accept run_tree to get
                    # around this.

                otel_context_manager = _maybe_create_otel_context(
                    run_container["new_run"]
                )

                async_gen_result = func(*args, **kwargs)
                # Can't iterate through if it's a coroutine
                accepts_context = aitertools.asyncio_accepts_context()
                use_ctx_task = accepts_context and not _runtime_override_active()
                if inspect.iscoroutine(async_gen_result):
                    if use_ctx_task:
                        async_gen_result = await asyncio.create_task(
                            async_gen_result, context=run_container["context"]
                        )  # type: ignore
                    else:
                        with tracing_context(
                            **get_tracing_context(run_container["context"])
                        ):
                            async_gen_result = await async_gen_result

                async for item in _process_async_iterator(
                    generator=async_gen_result,
                    run_container=run_container,
                    is_llm_run=(
                        run_container["new_run"].run_type == "llm"
                        if run_container["new_run"]
                        else False
                    ),
                    accepts_context=use_ctx_task,
                    results=results,
                    process_chunk=container_input.get("process_chunk"),
                    otel_context_manager=otel_context_manager,
                ):
                    yield item
            except BaseException as e:
                _cleanup_traceback(e)
                await asyncio.shield(
                    aitertools.aio_to_thread(
                        copy_context(),
                        _on_run_end,
                        run_container,
                        error=e,
                        outputs=_get_function_result(results, reduce_fn),
                    )
                )
                raise
            await aitertools.aio_to_thread(
                copy_context(),
                _on_run_end,
                run_container,
                outputs=_get_function_result(results, reduce_fn),
            )

        @functools.wraps(func)
        def wrapper(
            *args: Any,
            langsmith_extra: Optional[LangSmithExtra] = None,
            **kwargs: Any,
        ) -> Any:
            """Create a new run or create_child() if run is passed in kwargs."""
            if not func_accepts_config:
                kwargs.pop("config", None)
            run_container = _setup_run(
                func,
                container_input=container_input,
                langsmith_extra=langsmith_extra,
                args=args,
                kwargs=kwargs,
            )
            func_accepts_parent_run = (
                inspect.signature(func).parameters.get("run_tree", None) is not None
            )
            try:
                if func_accepts_parent_run:
                    kwargs["run_tree"] = run_container["new_run"]

                otel_context_manager = _maybe_create_otel_context(
                    run_container["new_run"]
                )
                if otel_context_manager:

                    def run_with_otel_context():
                        with otel_context_manager:
                            return func(*args, **kwargs)

                    function_result = run_container["context"].run(
                        run_with_otel_context
                    )
                else:
                    function_result = run_container["context"].run(
                        func, *args, **kwargs
                    )
            except BaseException as e:
                _cleanup_traceback(e)
                _on_run_end(run_container, error=e)
                raise
            _on_run_end(run_container, outputs=function_result)
            return function_result

        @functools.wraps(func)
        def generator_wrapper(
            *args: Any, langsmith_extra: Optional[LangSmithExtra] = None, **kwargs: Any
        ) -> Any:
            if not func_accepts_config:
                kwargs.pop("config", None)
            run_container = _setup_run(
                func,
                container_input=container_input,
                langsmith_extra=langsmith_extra,
                args=args,
                kwargs=kwargs,
            )
            func_accepts_parent_run = (
                inspect.signature(func).parameters.get("run_tree", None) is not None
            )
            results: list[Any] = []
            function_return: Any = None

            try:
                if func_accepts_parent_run:
                    kwargs["run_tree"] = run_container["new_run"]

                generator_result = run_container["context"].run(func, *args, **kwargs)

                otel_context_manager = _maybe_create_otel_context(
                    run_container["new_run"]
                )

                function_return = yield from _process_iterator(
                    generator_result,
                    run_container,
                    is_llm_run=run_type == "llm",
                    results=results,
                    process_chunk=container_input.get("process_chunk"),
                    otel_context_manager=otel_context_manager,
                )

                if function_return is not None:
                    results.append(function_return)

            except BaseException as e:
                _cleanup_traceback(e)
                _on_run_end(
                    run_container,
                    error=e,
                    outputs=_get_function_result(results, reduce_fn),
                )
                raise
            _on_run_end(run_container, outputs=_get_function_result(results, reduce_fn))

            return function_return

        # "Stream" functions (used in methods like OpenAI/Anthropic's SDKs)
        # are functions that return iterable responses and should not be
        # considered complete until the streaming is completed
        @functools.wraps(func)
        def stream_wrapper(
            *args: Any, langsmith_extra: Optional[LangSmithExtra] = None, **kwargs: Any
        ) -> Any:
            if not func_accepts_config:
                kwargs.pop("config", None)
            trace_container = _setup_run(
                func,
                container_input=container_input,
                langsmith_extra=langsmith_extra,
                args=args,
                kwargs=kwargs,
            )

            try:
                if func_accepts_parent_run:
                    kwargs["run_tree"] = trace_container["new_run"]
                stream = trace_container["context"].run(func, *args, **kwargs)
            except Exception as e:
                _cleanup_traceback(e)
                _on_run_end(trace_container, error=e)
                raise

            if hasattr(stream, "__iter__"):
                return _TracedStream(stream, trace_container, reduce_fn)
            elif hasattr(stream, "__aiter__"):
                # sync function -> async iterable (unexpected)
                return _TracedAsyncStream(stream, trace_container, reduce_fn)

            # If it's not iterable, end the trace immediately
            _on_run_end(trace_container, outputs=stream)
            return stream

        @functools.wraps(func)
        async def async_stream_wrapper(
            *args: Any, langsmith_extra: Optional[LangSmithExtra] = None, **kwargs: Any
        ) -> Any:
            if not func_accepts_config:
                kwargs.pop("config", None)
            trace_container = await aitertools.aio_to_thread(
                copy_context(),
                _setup_run,
                func,
                container_input=container_input,
                langsmith_extra=langsmith_extra,
                args=args,
                kwargs=kwargs,
            )

            try:
                if func_accepts_parent_run:
                    kwargs["run_tree"] = trace_container["new_run"]
                stream = await func(*args, **kwargs)
            except Exception as e:
                await aitertools.aio_to_thread(
                    copy_context(), _on_run_end, trace_container, error=e
                )
                raise

            if hasattr(stream, "__aiter__"):
                return _TracedAsyncStream(stream, trace_container, reduce_fn)
            elif hasattr(stream, "__iter__"):
                # Async function -> sync iterable
                return _TracedStream(stream, trace_container, reduce_fn)

            # If it's not iterable, end the trace immediately
            await aitertools.aio_to_thread(
                copy_context(), _on_run_end, trace_container, outputs=stream
            )
            return stream

        if inspect.isasyncgenfunction(func):
            selected_wrapper: Callable = async_generator_wrapper
        elif inspect.isgeneratorfunction(func):
            selected_wrapper = generator_wrapper
        elif is_async(func):
            if reduce_fn:
                selected_wrapper = async_stream_wrapper
            else:
                selected_wrapper = async_wrapper
        else:
            if reduce_fn:
                selected_wrapper = stream_wrapper
            else:
                selected_wrapper = wrapper
        setattr(selected_wrapper, "__langsmith_traceable__", True)
        setattr(
            selected_wrapper,
            "__traceable_config__",
            {
                "process_inputs": container_input.get("process_inputs"),
                "process_outputs": outputs_processor,
                "enabled": enabled,
                "tags": container_input.get("tags"),
                "metadata": container_input.get("metadata"),
                "wrapped": func,
            },
        )
        sig = inspect.signature(selected_wrapper)
        if not sig.parameters.get("config"):
            sig = sig.replace(
                parameters=[
                    *(
                        param
                        for param in sig.parameters.values()
                        if param.kind != inspect.Parameter.VAR_KEYWORD
                    ),
                    inspect.Parameter(
                        "config", inspect.Parameter.KEYWORD_ONLY, default=None
                    ),
                    *(
                        param
                        for param in sig.parameters.values()
                        if param.kind == inspect.Parameter.VAR_KEYWORD
                    ),
                ]
            )
            selected_wrapper.__signature__ = sig  # type: ignore[attr-defined]
        return selected_wrapper

    # If the decorator is called with no arguments, then it's being used as a
    # decorator, so we return the decorator function
    if len(args) == 1 and callable(args[0]) and not kwargs:
        return decorator(args[0])
    # Else it's being used as a decorator factory, so we return the decorator
    return decorator


class trace:
    """Manage a LangSmith run in context.

    This class can be used as both a synchronous and asynchronous context manager.

    Args:
        name: Name of the run.
        run_type: Type of run (e.g., `'chain'`, `'llm'`, `'tool'`).
        inputs: Initial input data for the run.
        project_name: Project name to associate the run with.
        parent: Parent run.

            Can be a `RunTree`, dotted order string, or tracing headers.
        tags: List of tags for the run.
        metadata: Additional metadata for the run.
        client: LangSmith client for custom settings.
        run_id: Preset identifier for the run.
        reference_example_id: Associates run with a dataset example.

            Only for root runs in evaluation.
        exceptions_to_handle: Exception types to ignore.
        extra: Extra data to send to LangSmith.

            Use 'metadata' instead.

    Examples:
        Synchronous usage:

        ```python
        with trace("My Operation", run_type="tool", tags=["important"]) as run:
            result = "foo"  # Perform operation
            run.metadata["some-key"] = "some-value"
            run.end(outputs={"result": result})
        ```

        Asynchronous usage:

        ```python
        async def main():
            async with trace("Async Operation", run_type="tool", tags=["async"]) as run:
                result = "foo"  # Await async operation
                run.metadata["some-key"] = "some-value"
                # "end" just adds the outputs and sets error to None
                # The actual patching of the run happens when the context exits
                run.end(outputs={"result": result})


        asyncio.run(main())
        ```

        Handling specific exceptions:

        ```python
        import pytest
        import sys

        with trace("Test", exceptions_to_handle=(pytest.skip.Exception,)):
            if sys.platform == "win32":  # Just an example
                pytest.skip("Skipping test for windows")
            result = "foo"  # Perform test operation
        ```
    """

    def __init__(
        self,
        name: str,
        run_type: ls_client.RUN_TYPE_T = "chain",
        *,
        inputs: Optional[dict] = None,
        extra: Optional[dict] = None,
        project_name: Optional[str] = None,
        parent: Optional[
            Union[run_trees.RunTree, str, Mapping, Literal["ignore"]]
        ] = None,
        tags: Optional[list[str]] = None,
        metadata: Optional[Mapping[str, Any]] = None,
        client: Optional[ls_client.Client] = None,
        run_id: Optional[ls_client.ID_TYPE] = None,
        reference_example_id: Optional[ls_client.ID_TYPE] = None,
        exceptions_to_handle: Optional[tuple[type[BaseException], ...]] = None,
        attachments: Optional[schemas.Attachments] = None,
        **kwargs: Any,
    ):
        """Initialize the trace context manager.

        Warns if unsupported kwargs are passed.
        """
        self._end_on_exit = kwargs.pop("_end_on_exit", True)
        if kwargs:
            warnings.warn(
                "The `trace` context manager no longer supports the following kwargs: "
                f"{sorted(kwargs.keys())}.",
                DeprecationWarning,
            )
        self.name = name
        self.run_type = run_type
        self.inputs = inputs
        self.attachments = attachments
        self.extra = extra
        self.project_name = project_name
        self.parent = parent
        # The run tree is deprecated. Keeping for backwards compat.
        # Will fully merge within parent later.
        self.run_tree = kwargs.get("run_tree")
        self.tags = tags
        self.metadata = metadata
        self.client = client
        self.run_id = run_id
        self.reference_example_id = reference_example_id
        self.exceptions_to_handle = exceptions_to_handle
        self.new_run: Optional[run_trees.RunTree] = None
        self.old_ctx: Optional[dict] = None

    def _setup(self) -> run_trees.RunTree:
        """Set up the tracing context and create a new run.

        This method initializes the tracing context, merges tags and metadata,
        creates a new run (either as a child of an existing run or as a new root run),
        and sets up the necessary context variables.

        Returns:
            run_trees.RunTree: The newly created run.
        """
        self.old_ctx = get_tracing_context()
        enabled = utils.tracing_is_enabled(self.old_ctx)

        outer_tags = _context._TAGS.get() or _context._GLOBAL_TAGS
        outer_metadata = _context._METADATA.get() or _context._GLOBAL_METADATA
        client_ = self.client or self.old_ctx.get("client")
        parent_run_ = _get_parent_run(
            {
                "parent": self.parent,
                "run_tree": self.run_tree,
                "client": client_,
                "project_name": self.project_name,
            }
        )

        tags_ = sorted(set((self.tags or []) + (outer_tags or [])))
        metadata = {
            **(self.metadata or {}),
            **(outer_metadata or {}),
            "ls_method": "trace",
        }

        extra_outer = self.extra or {}
        extra_outer["metadata"] = metadata

        project_name_ = _get_project_name(self.project_name)

        if parent_run_ is not None and enabled:
            self.new_run = parent_run_.create_child(
                name=self.name,
                run_id=self.run_id,
                run_type=self.run_type,
                extra=extra_outer,
                inputs=self.inputs,
                tags=tags_,
                attachments=self.attachments,
            )
        else:
            self.new_run = run_trees.RunTree(
                name=self.name,
                id=ls_client._ensure_uuid(self.run_id),
                reference_example_id=ls_client._ensure_uuid(
                    self.reference_example_id, accept_null=True
                ),
                run_type=self.run_type,
                extra=extra_outer,
                project_name=project_name_ or "default",
                replicas=run_trees._REPLICAS.get(),
                inputs=self.inputs or {},
                tags=tags_,
                client=client_,  # type: ignore
                attachments=self.attachments or {},  # type: ignore
            )

        if enabled is True:
            self.new_run.post()
        if enabled:
            _context._TAGS.set(tags_)
            _context._METADATA.set(metadata)
            if self.new_run is not None:
                _context._PARENT_RUN_TREE_REF.set(weakref.ref(self.new_run))
            else:
                _context._PARENT_RUN_TREE_REF.set(None)
            _context._PROJECT_NAME.set(project_name_)
            _context._CLIENT.set(client_)

        return self.new_run

    def _teardown(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Clean up the tracing context and finalize the run.

        This method handles exceptions, ends the run if necessary,
        patches the run if it's not disabled, and resets the tracing context.

        Args:
            exc_type: The type of the exception that occurred, if any.
            exc_value: The exception instance that occurred, if any.
            traceback: The traceback object associated with the exception, if any.
        """
        if self.new_run is None:
            return
        if exc_type is not None:
            tb = _format_error_with_exceptions_to_handle(
                exc_value, self.exceptions_to_handle
            )
            self.new_run.end(error=tb)
        if self.old_ctx is not None:
            enabled = utils.tracing_is_enabled(self.old_ctx)
            if enabled is True and self._end_on_exit:
                self.new_run.patch()

            _set_tracing_context(self.old_ctx)
        else:
            warnings.warn("Tracing context was not set up properly.", RuntimeWarning)

    def __enter__(self) -> run_trees.RunTree:
        """Enter the context manager synchronously.

        Returns:
            run_trees.RunTree: The newly created run.
        """
        return self._setup()

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        """Exit the context manager synchronously.

        Args:
            exc_type: The type of the exception that occurred, if any.
            exc_value: The exception instance that occurred, if any.
            traceback: The traceback object associated with the exception, if any.
        """
        self._teardown(exc_type, exc_value, traceback)

    async def __aenter__(self) -> run_trees.RunTree:
        """Enter the context manager asynchronously.

        Returns:
            run_trees.RunTree: The newly created run.
        """
        ctx = copy_context()
        result = await aitertools.aio_to_thread(ctx, self._setup)
        # Set the context for the current thread
        _set_tracing_context(get_tracing_context(ctx))
        return result

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        """Exit the context manager asynchronously.

        Args:
            exc_type: The type of the exception that occurred, if any.
            exc_value: The exception instance that occurred, if any.
            traceback: The traceback object associated with the exception, if any.
        """
        ctx = copy_context()
        if exc_type is not None:
            await asyncio.shield(
                aitertools.aio_to_thread(
                    ctx, self._teardown, exc_type, exc_value, traceback
                )
            )
        else:
            await aitertools.aio_to_thread(
                ctx, self._teardown, exc_type, exc_value, traceback
            )
        _set_tracing_context(get_tracing_context(ctx))


def _get_project_name(project_name: Optional[str]) -> Optional[str]:
    if project_name:
        return project_name
    prt = get_current_run_tree()
    return (
        # Maintain tree consistency first
        _context._PROJECT_NAME.get()
        or (prt.session_name if prt else None)
        # Global fallback configured via ls.configure(...)
        or _context._GLOBAL_PROJECT_NAME
        # fallback to the default for the environment
        or utils.get_tracer_project()
    )


def as_runnable(traceable_fn: Callable) -> Runnable:
    """Convert a function wrapped by the LangSmith `@traceable` decorator to a `Runnable`.

    Args:
        traceable_fn: The function wrapped by the `@traceable` decorator.

    Returns:
        Runnable: A `Runnable` object that maintains a consistent LangSmith
            tracing context.

    Raises:
        ImportError: If `langchain` module is not installed.
        ValueError: If the provided function is not wrapped by the `@traceable` decorator.

    Example:
        >>> @traceable
        ... def my_function(input_data):
        ...     # Function implementation
        ...     pass
        >>> runnable = as_runnable(my_function)
    """
    try:
        from langchain_core.runnables import RunnableConfig, RunnableLambda
        from langchain_core.runnables.utils import Input, Output
    except ImportError as e:
        raise ImportError(
            "as_runnable requires langchain-core to be installed. "
            "You can install it with `pip install langchain-core`."
        ) from e
    if not is_traceable_function(traceable_fn):
        try:
            fn_src = inspect.getsource(traceable_fn)
        except Exception:
            fn_src = "<source unavailable>"
        raise ValueError(
            f"as_runnable expects a function wrapped by the LangSmith"
            f" @traceable decorator. Got {traceable_fn} defined as:\n{fn_src}"
        )

    class RunnableTraceable(RunnableLambda):
        """Converts a `@traceable` decorated function to a `Runnable`.

        This helps maintain a consistent LangSmith tracing context.
        """

        def __init__(
            self,
            func: Callable,
            afunc: Optional[Callable[..., Awaitable[Output]]] = None,
        ) -> None:
            wrapped: Optional[Callable[[Input], Output]] = None
            awrapped = self._wrap_async(afunc)
            if is_async(func):
                if awrapped is not None:
                    raise TypeError(
                        "Func was provided as a coroutine function, but afunc was "
                        "also provided. If providing both, func should be a regular "
                        "function to avoid ambiguity."
                    )
                wrapped = cast(Callable[[Input], Output], self._wrap_async(func))
            elif is_traceable_function(func):
                wrapped = cast(Callable[[Input], Output], self._wrap_sync(func))
            if wrapped is None:
                raise ValueError(
                    f"{self.__class__.__name__} expects a function wrapped by"
                    " the LangSmith"
                    f" @traceable decorator. Got {func}"
                )

            super().__init__(
                wrapped,
                cast(
                    Optional[Callable[[Input], Awaitable[Output]]],
                    awrapped,
                ),
            )

        @staticmethod
        def _wrap_sync(
            func: Callable[..., Output],
        ) -> Callable[[Input, RunnableConfig], Output]:
            """Wrap a synchronous function to make it asynchronous."""

            def wrap_traceable(inputs: dict, config: RunnableConfig) -> Any:
                run_tree = run_trees.RunTree.from_runnable_config(cast(dict, config))
                return func(**inputs, langsmith_extra={"run_tree": run_tree})

            return cast(Callable[[Input, RunnableConfig], Output], wrap_traceable)

        @staticmethod
        def _wrap_async(
            afunc: Optional[Callable[..., Awaitable[Output]]],
        ) -> Optional[Callable[[Input, RunnableConfig], Awaitable[Output]]]:
            """Wrap an async function to make it synchronous."""
            if afunc is None:
                return None

            if not is_traceable_function(afunc):
                raise ValueError(
                    "RunnableTraceable expects a function wrapped by the LangSmith"
                    f" @traceable decorator. Got {afunc}"
                )
            afunc_ = cast(Callable[..., Awaitable[Output]], afunc)

            async def awrap_traceable(inputs: dict, config: RunnableConfig) -> Any:
                run_tree = run_trees.RunTree.from_runnable_config(cast(dict, config))
                return await afunc_(**inputs, langsmith_extra={"run_tree": run_tree})

            return cast(
                Callable[[Input, RunnableConfig], Awaitable[Output]], awrap_traceable
            )

    return RunnableTraceable(traceable_fn)


## Private Methods and Objects
_VALID_RUN_TYPES = {
    "tool",
    "chain",
    "llm",
    "retriever",
    "embedding",
    "prompt",
    "parser",
}


class _TraceableContainer(TypedDict, total=False):
    """Typed response when initializing a run a traceable."""

    new_run: Optional[run_trees.RunTree]
    project_name: Optional[str]
    outer_project: Optional[str]
    outer_metadata: Optional[dict[str, Any]]
    outer_tags: Optional[list[str]]
    _on_success: Optional[Callable[[run_trees.RunTree], Any]]
    on_end: Optional[Callable[[run_trees.RunTree], Any]]
    context: contextvars.Context
    _token_event_logged: Optional[bool]
    enabled: Optional[bool]
    exceptions_to_handle: Optional[tuple[type[BaseException], ...]]


class _ContainerInput(TypedDict, total=False):
    """Typed response when initializing a run a traceable."""

    extra_outer: Optional[dict]
    name: Optional[str]
    metadata: Optional[dict[str, Any]]
    tags: Optional[list[str]]
    client: Optional[ls_client.Client]
    reduce_fn: Optional[Callable]
    project_name: Optional[str]
    run_type: ls_client.RUN_TYPE_T
    process_inputs: Optional[Callable[[dict], dict]]
    process_chunk: Optional[Callable]
    invocation_params_fn: Optional[Callable[[dict], dict]]
    dangerously_allow_filesystem: Optional[bool]
    enabled: Optional[bool]
    exceptions_to_handle: Optional[tuple[type[BaseException], ...]]


def _format_error_with_exceptions_to_handle(
    error: BaseException | None,
    exceptions_to_handle: Optional[tuple[type[BaseException], ...]],
) -> Optional[str]:
    """Format error for logging, suppressing traceback for handled exceptions."""
    if exceptions_to_handle and isinstance(error, exceptions_to_handle):
        return None

    stacktrace = utils._format_exc()
    return f"{repr(error)}\n\n{stacktrace}"


def _container_end(
    container: _TraceableContainer,
    outputs: Optional[Any] = None,
    error: Optional[BaseException] = None,
) -> None:
    """End the run."""
    run_tree = container.get("new_run")
    if run_tree is None:
        # Tracing not enabled
        return
    if isinstance(outputs, dict):
        dict_outputs = outputs
    elif (
        outputs is not None
        and hasattr(outputs, "model_dump")
        and callable(outputs.model_dump)
        and not isinstance(outputs, type)
    ):
        try:
            dict_outputs = outputs.model_dump(exclude_none=True, mode="json")
        except Exception as e:
            LOGGER.debug(
                f"Failed to use model_dump to serialize {type(outputs)} to JSON: {e}"
            )
            dict_outputs = {"output": outputs}
    else:
        dict_outputs = {"output": outputs}
    if (usage := _extract_usage(run_tree=run_tree, outputs=dict_outputs)) is not None:
        run_tree.metadata["usage_metadata"] = usage
    if error:
        error_repr = _format_error_with_exceptions_to_handle(
            error, container.get("exceptions_to_handle")
        )
    else:
        error_repr = None
        if (_on_success := container.get("_on_success")) and callable(_on_success):
            try:
                _on_success(run_tree)
            except BaseException as e:
                warnings.warn(f"Failed to run _on_success function: {e}")

    run_tree.end(outputs=dict_outputs, error=error_repr)
    # Patch run if enabled=True (force) or if tracing is enabled globally
    enabled = container.get("enabled")
    if enabled is True or utils.tracing_is_enabled() is True:
        run_tree.patch()
    if (on_end := container.get("on_end")) and callable(on_end):
        try:
            on_end(run_tree)
        except BaseException as e:
            warnings.warn(f"Failed to run on_end function: {e}")


def _collect_extra(extra_outer: dict, langsmith_extra: LangSmithExtra) -> dict:
    run_extra = langsmith_extra.get("run_extra", None)
    if run_extra:
        extra_inner = {**extra_outer, **run_extra}
    else:
        extra_inner = extra_outer
    return extra_inner


def _get_parent_run(
    langsmith_extra: LangSmithExtra,
    config: Optional[dict] = None,
) -> Optional[run_trees.RunTree]:
    parent = langsmith_extra.get("parent")
    if parent == "ignore":
        return None
    if isinstance(parent, run_trees.RunTree):
        return parent
    if isinstance(parent, Mapping):
        return run_trees.RunTree.from_headers(
            parent,
            client=langsmith_extra.get("client"),
            # Precedence: headers -> cvar -> explicit -> env var
            project_name=_get_project_name(langsmith_extra.get("project_name")),
        )
    if isinstance(parent, str):
        dort = run_trees.RunTree.from_dotted_order(
            parent,
            client=langsmith_extra.get("client"),
            # Precedence: cvar -> explicit ->  env var
            project_name=_get_project_name(langsmith_extra.get("project_name")),
        )
        return dort
    run_tree = langsmith_extra.get("run_tree")
    if run_tree:
        return run_tree
    crt = get_current_run_tree()
    if _runtime_env.get_langchain_core_version() is not None:
        if rt := run_trees.RunTree.from_runnable_config(
            config, client=langsmith_extra.get("client")
        ):
            # Still need to break ties when alternating between traceable and
            # LanChain code.
            # Nesting: LC -> LS -> LS, we want to still use LS as the parent
            # Otherwise would look like LC -> {LS, LS} (siblings)
            if (
                not crt  # Simple LC -> LS
                # Let user override if manually passed in or invoked in a
                # RunnableSequence. This is a naive check.
                or (config is not None and config.get("callbacks"))
                # If the LangChain dotted order is more nested than the LangSmith
                # dotted order, use the LangChain run as the parent.
                # Note that this condition shouldn't be triggered in later
                # versions of core, since we also update the run_tree context
                # vars when updating the RunnableConfig context var.
                or rt.dotted_order > crt.dotted_order
            ):
                return rt
    return crt


def _setup_run(
    func: Callable,
    container_input: _ContainerInput,
    langsmith_extra: Optional[LangSmithExtra] = None,
    args: Any = None,
    kwargs: Any = None,
) -> _TraceableContainer:
    """Create a new run or create_child() if run is passed in kwargs."""
    extra_outer = container_input.get("extra_outer") or {}
    metadata = container_input.get("metadata")
    tags = container_input.get("tags")
    client = container_input.get("client")
    run_type = container_input.get("run_type") or "chain"
    dangerously_allow_filesystem = container_input.get(
        "dangerously_allow_filesystem", False
    )
    outer_project = _context._PROJECT_NAME.get()
    langsmith_extra = langsmith_extra or LangSmithExtra()
    name = langsmith_extra.get("name") or container_input.get("name")
    client_ = langsmith_extra.get("client", client) or _context._CLIENT.get()
    parent_run_ = _get_parent_run(
        {**langsmith_extra, "client": client_}, kwargs.get("config")
    )
    project_cv = _context._PROJECT_NAME.get()
    selected_project = (
        project_cv  # From parent trace
        or (
            parent_run_.session_name if parent_run_ else None
        )  # from parent run attempt 2 (not managed by traceable)
        or langsmith_extra.get("project_name")  # at invocation time
        or container_input["project_name"]  # at decorator time
        or _context._GLOBAL_PROJECT_NAME  # global fallback from ls.configure
        or utils.get_tracer_project()  # default
    )
    reference_example_id = langsmith_extra.get("reference_example_id")
    id_ = langsmith_extra.get("run_id")
    enabled = container_input.get("enabled")
    # Determine if tracing should be enabled for this function:
    # - enabled=False: never trace
    # - enabled=True: always trace
    # - enabled=None: use context/environment setting
    if enabled is False or (
        enabled is not True and not (parent_run_ or utils.tracing_is_enabled())
    ):
        utils.log_once(
            logging.DEBUG,
            "LangSmith tracing is not enabled, returning original function.",
        )
        context = copy_context()
        # Clear the parent RunTree context so nested calls don't inherit it
        context.run(_context._PARENT_RUN_TREE_REF.set, None)
        return _TraceableContainer(
            new_run=None,
            project_name=selected_project,
            outer_project=outer_project,
            outer_metadata=None,
            outer_tags=None,
            _on_success=langsmith_extra.get("_on_success"),
            on_end=langsmith_extra.get("on_end"),
            context=context,
            _token_event_logged=False,
            exceptions_to_handle=container_input.get("exceptions_to_handle"),
        )
    signature = inspect.signature(func)
    name_ = name or utils._get_function_name(func)
    extra_inner = _collect_extra(extra_outer, langsmith_extra)
    outer_metadata = _context._METADATA.get() or _context._GLOBAL_METADATA
    outer_tags = _context._TAGS.get() or _context._GLOBAL_TAGS
    context = copy_context()
    metadata_ = {
        **(langsmith_extra.get("metadata") or {}),
        **(outer_metadata or {}),
    }
    context.run(_context._METADATA.set, metadata_)
    metadata_.update(metadata or {})
    metadata_["ls_method"] = "traceable"
    extra_inner["metadata"] = metadata_
    inputs, attachments = _get_inputs_and_attachments_safe(
        signature, *args, func=func, **kwargs
    )
    invocation_params_fn = container_input.get("invocation_params_fn")
    if invocation_params_fn:
        try:
            invocation_params = {
                k: v for k, v in invocation_params_fn(inputs).items() if v is not None
            }
            if invocation_params and isinstance(invocation_params, dict):
                metadata_.update(invocation_params)
        except BaseException as e:
            LOGGER.error(f"Failed to infer invocation params for {name_}: {e}")
    process_inputs = container_input.get("process_inputs")
    if process_inputs:
        try:
            inputs = process_inputs(inputs)
        except BaseException as e:
            LOGGER.error(f"Failed to filter inputs for {name_}: {e}")
    tags_ = (langsmith_extra.get("tags") or []) + (outer_tags or [])
    context.run(_context._TAGS.set, tags_)
    tags_ += tags or []
    if parent_run_ is not None:
        new_run = parent_run_.create_child(
            name=name_,
            run_type=run_type,
            inputs=inputs,
            tags=tags_,
            extra=extra_inner,
            run_id=id_,
            attachments=attachments,
        )
    else:
        # Create RunTree kwargs conditionally to let RunTree generate id from start_time
        run_tree_kwargs = {
            "name": name_,
            "inputs": inputs,
            "run_type": run_type,
            "reference_example_id": ls_client._ensure_uuid(
                reference_example_id, accept_null=True
            ),
            "project_name": selected_project,
            "replicas": run_trees._REPLICAS.get(),
            "extra": extra_inner,
            "tags": tags_,
            "client": client_,
            "attachments": attachments,
            "dangerously_allow_filesystem": dangerously_allow_filesystem,
        }
        # Only pass id if user explicitly provided one
        if id_ is not None:
            run_tree_kwargs["id"] = ls_client._ensure_uuid(id_)
        new_run = run_trees.RunTree(**cast(Any, run_tree_kwargs))
    # Post run if enabled=True (force) or if tracing is enabled globally
    if enabled is True or utils.tracing_is_enabled() is True:
        try:
            new_run.post()
        except BaseException as e:
            LOGGER.error(f"Failed to post run {new_run.id}: {e}")
    response_container = _TraceableContainer(
        new_run=new_run,
        project_name=selected_project,
        outer_project=outer_project,
        outer_metadata=outer_metadata,
        outer_tags=outer_tags,
        on_end=langsmith_extra.get("on_end"),
        _on_success=langsmith_extra.get("_on_success"),
        context=context,
        _token_event_logged=False,
        enabled=enabled,
        exceptions_to_handle=container_input.get("exceptions_to_handle"),
    )
    context.run(_context._PROJECT_NAME.set, response_container["project_name"])
    # Store a weak reference (not the RunTree itself) in the copied context.
    # This prevents memory leaks when contexts are captured by asyncio operations
    # (call_later, create_task, etc.) — the captured context holds only a weakref,
    # and the RunTree can be GC'd once no strong references remain.
    if new_run is not None:
        context.run(_context._PARENT_RUN_TREE_REF.set, weakref.ref(new_run))
    else:
        context.run(_context._PARENT_RUN_TREE_REF.set, None)
    return response_container


def _handle_container_end(
    container: _TraceableContainer,
    outputs: Optional[Any] = None,
    error: Optional[BaseException] = None,
    outputs_processor: Optional[Callable[..., dict]] = None,
) -> None:
    """Handle the end of run."""
    try:
        if outputs_processor is not None:
            outputs = outputs_processor(outputs)
        _container_end(container, outputs=outputs, error=error)
    except BaseException as e:
        LOGGER.warning(f"Unable to process trace outputs: {repr(e)}")


def _is_traceable_function(func: Any) -> bool:
    return getattr(func, "__langsmith_traceable__", False)


def _get_inputs(
    signature: inspect.Signature, *args: Any, **kwargs: Any
) -> dict[str, Any]:
    """Return a dictionary of inputs from the function signature."""
    bound = signature.bind_partial(*args, **kwargs)
    bound.apply_defaults()
    arguments = dict(bound.arguments)
    arguments.pop("self", None)
    arguments.pop("cls", None)
    for param_name, param in signature.parameters.items():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            # Update with the **kwargs, and remove the original entry
            # This is to help flatten out keyword arguments
            if param_name in arguments:
                arguments.update(arguments[param_name])
                arguments.pop(param_name)

    return arguments


def _get_inputs_safe(
    signature: inspect.Signature, *args: Any, **kwargs: Any
) -> dict[str, Any]:
    try:
        return _get_inputs(signature, *args, **kwargs)
    except BaseException as e:
        LOGGER.debug(f"Failed to get inputs for {signature}: {e}")
        return {"args": args, "kwargs": kwargs}


def _is_attachment(param: inspect.Parameter, func: Optional[Callable] = None) -> bool:
    if param.annotation == schemas.Attachment or (
        get_origin(param.annotation) == Annotated
        and any(arg == schemas.Attachment for arg in get_args(param.annotation))
    ):
        return True

    # try resolving stringified annotations
    if func is not None and isinstance(param.annotation, str):
        try:
            # include_extras=True preserves annotated metadata
            type_hints = get_type_hints(func, include_extras=True)
            resolved_annotation = type_hints.get(param.name)
            if resolved_annotation is not None:
                return resolved_annotation == schemas.Attachment or (
                    get_origin(resolved_annotation) == Annotated
                    and any(
                        arg == schemas.Attachment
                        for arg in get_args(resolved_annotation)
                    )
                )
        except (NameError, TypeError, AttributeError):
            pass

    return False


def _attachment_args_helper(
    signature: inspect.Signature, func: Optional[Callable] = None
) -> set[str]:
    return {
        name
        for name, param in signature.parameters.items()
        if _is_attachment(param, func)
    }


@functools.lru_cache(maxsize=1000)
def _cached_attachment_args_by_signature(signature: inspect.Signature) -> set[str]:
    return _attachment_args_helper(signature, None)


_attachment_args_cache: weakref.WeakKeyDictionary[
    Callable[..., Any], tuple[inspect.Signature, set[str]]
] = weakref.WeakKeyDictionary()
_attachment_args_cache_lock = threading.Lock()


def _cached_attachment_args(
    signature: inspect.Signature, func: Optional[Callable] = None
) -> set[str]:
    if func is None:
        return _cached_attachment_args_by_signature(signature)

    try:
        with _attachment_args_cache_lock:
            cached = _attachment_args_cache.get(func)
    except TypeError:
        # Some callable objects cannot be weak-referenced.
        return _attachment_args_helper(signature, func)

    if cached is not None and cached[0] == signature:
        return cached[1]

    result = _attachment_args_helper(signature, func)
    with _attachment_args_cache_lock:
        _attachment_args_cache[func] = (signature, result)
    return result


def _clear_cached_attachment_args() -> None:
    with _attachment_args_cache_lock:
        _attachment_args_cache.clear()
    _cached_attachment_args_by_signature.cache_clear()


_cached_attachment_args.cache_clear = _clear_cached_attachment_args  # type: ignore[attr-defined]


def _attachment_args(
    signature: inspect.Signature, func: Optional[Callable] = None
) -> set[str]:
    # Caching signatures fails if there's unhashable default values.
    try:
        return _cached_attachment_args(signature, func)
    except TypeError:
        return _attachment_args_helper(signature, func)


def _get_inputs_and_attachments_safe(
    signature: inspect.Signature,
    *args: Any,
    func: Optional[Callable] = None,
    **kwargs: Any,
) -> tuple[dict, schemas.Attachments]:
    try:
        inferred = _get_inputs(signature, *args, **kwargs)
        attachment_args = _attachment_args(signature, func)
        if attachment_args:
            inputs, attachments = {}, {}
            for k, v in inferred.items():
                if k in attachment_args:
                    attachments[k] = v
                else:
                    inputs[k] = v
            return inputs, attachments
        return inferred, {}
    except BaseException as e:
        LOGGER.warning(f"Failed to get inputs for {signature}: {e}")
        return {"args": args, "kwargs": kwargs}, {}


def _set_tracing_context(context: Optional[dict[str, Any]] = None):
    """Set the tracing context."""
    if context is None:
        for k, v in _CONTEXT_KEYS.items():
            v.set(None)
        return
    for k, v in context.items():
        if k == "parent":
            # "parent" is a RunTree — store a weak reference
            if v is not None:
                _context._PARENT_RUN_TREE_REF.set(weakref.ref(v))
            else:
                _context._PARENT_RUN_TREE_REF.set(None)
            continue
        var = _CONTEXT_KEYS.get(k)
        if var is not None:
            var.set(v)


def _process_iterator(
    generator: Iterator[T],
    run_container: _TraceableContainer,
    is_llm_run: bool,
    # Results is mutated
    results: list[Any],
    process_chunk: Optional[Callable],
    otel_context_manager: Optional[Any] = None,
) -> Generator[T, None, Any]:
    try:
        while True:
            if otel_context_manager:
                # Create a fresh context manager for each iteration
                def next_with_otel_context():
                    # Get the run_tree from run_container to create a fresh context
                    run_tree = run_container.get("new_run")
                    if run_tree:
                        fresh_otel_context = _maybe_create_otel_context(run_tree)
                        if fresh_otel_context:
                            with fresh_otel_context:
                                return next(generator)
                    return next(generator)

                item: T = run_container["context"].run(next_with_otel_context)  # type: ignore[arg-type]
            else:
                item = run_container["context"].run(next, generator)  # type: ignore[arg-type]
            if process_chunk:
                traced_item = process_chunk(item)
            else:
                traced_item = item
            if (
                is_llm_run
                and run_container["new_run"]
                and not run_container.get("_token_event_logged")
            ):
                run_container["new_run"].add_event(
                    {
                        "name": "new_token",
                        "time": datetime.datetime.now(
                            datetime.timezone.utc
                        ).isoformat(),
                        "kwargs": {"token": traced_item},
                    }
                )
                run_container["_token_event_logged"] = True
            results.append(traced_item)
            yield item
    except StopIteration as e:
        return e.value


async def _process_async_iterator(
    generator: AsyncIterator[T],
    run_container: _TraceableContainer,
    *,
    is_llm_run: bool,
    accepts_context: bool,
    results: list[Any],
    process_chunk: Optional[Callable],
    otel_context_manager: Optional[Any] = None,
) -> AsyncGenerator[T, None]:
    try:
        while True:
            if otel_context_manager:
                # Create a fresh context manager for each iteration
                async def anext_with_otel_context():
                    # Get the run_tree from run_container to create a fresh context
                    run_tree = run_container.get("new_run")
                    if run_tree:
                        fresh_otel_context = _maybe_create_otel_context(run_tree)
                        if fresh_otel_context:
                            with fresh_otel_context:
                                return await aitertools.py_anext(generator)
                    return await aitertools.py_anext(generator)

                if accepts_context:
                    item = await asyncio.create_task(  # type: ignore[call-arg, var-annotated]
                        anext_with_otel_context(),
                        context=run_container["context"],
                    )
                else:
                    # Python < 3.11
                    with tracing_context(
                        **get_tracing_context(run_container["context"])
                    ):
                        item = await anext_with_otel_context()
            else:
                if accepts_context:
                    item = await asyncio.create_task(  # type: ignore[call-arg, var-annotated]
                        aitertools.py_anext(generator),  # type: ignore[arg-type]
                        context=run_container["context"],
                    )
                else:
                    # Python < 3.11
                    with tracing_context(
                        **get_tracing_context(run_container["context"])
                    ):
                        item = await aitertools.py_anext(generator)
            if process_chunk:
                traced_item = process_chunk(item)
            else:
                traced_item = item
            if (
                is_llm_run
                and run_container["new_run"]
                and not run_container.get("_token_event_logged")
            ):
                run_container["new_run"].add_event(
                    {
                        "name": "new_token",
                        "time": datetime.datetime.now(
                            datetime.timezone.utc
                        ).isoformat(),
                        "kwargs": {"token": traced_item},
                    }
                )
                run_container["_token_event_logged"] = True
            results.append(traced_item)
            yield item
    except StopAsyncIteration:
        pass


T = TypeVar("T")


class _TracedStreamBase(Generic[T]):
    """Base class for traced stream objects."""

    def __init__(
        self,
        stream: Union[Iterator[T], AsyncIterator[T]],
        trace_container: _TraceableContainer,
        reduce_fn: Optional[Callable] = None,
    ):
        self.__ls_stream__ = stream
        self.__ls_trace_container__ = trace_container
        self.__ls_completed__ = False
        self.__ls_reduce_fn__ = reduce_fn
        self.__ls_accumulated_output__: list[T] = []
        self.__is_llm_run__ = (
            trace_container["new_run"].run_type == "llm"
            if trace_container["new_run"]
            else False
        )

    def __getattr__(self, name: str):
        return getattr(self.__ls_stream__, name)

    def __dir__(self):
        return list(set(dir(self.__class__) + dir(self.__ls_stream__)))

    def __repr__(self):
        return f"Traceable({self.__ls_stream__!r})"

    def __str__(self):
        return str(self.__ls_stream__)

    def __del__(self):
        try:
            if not self.__ls_completed__:
                self._end_trace()
        except BaseException:
            pass
        try:
            self.__ls_stream__.__del__()
        except BaseException:
            pass

    def _end_trace(self, error: Optional[BaseException] = None):
        if self.__ls_completed__:
            return
        try:
            if self.__ls_reduce_fn__:
                reduced_output = self.__ls_reduce_fn__(self.__ls_accumulated_output__)
            else:
                reduced_output = self.__ls_accumulated_output__
            _container_end(
                self.__ls_trace_container__,
                outputs=reduced_output,
                error=error,
            )
        finally:
            self.__ls_completed__ = True


class _TracedStream(_TracedStreamBase, Generic[T]):
    """A wrapper for synchronous stream objects that handles tracing."""

    def __init__(
        self,
        stream: Iterator[T],
        trace_container: _TraceableContainer,
        reduce_fn: Optional[Callable] = None,
        process_chunk: Optional[Callable] = None,
    ):
        super().__init__(
            stream=stream,
            trace_container=trace_container,
            reduce_fn=reduce_fn,
        )
        self.__ls_stream__ = stream
        self.__ls__gen__ = _process_iterator(
            self.__ls_stream__,
            self.__ls_trace_container__,
            is_llm_run=self.__is_llm_run__,
            results=self.__ls_accumulated_output__,
            process_chunk=process_chunk,
        )

    def __next__(self) -> T:
        try:
            return next(self.__ls__gen__)
        except StopIteration:
            self._end_trace()
            raise

    def __iter__(self) -> Iterator[T]:
        try:
            yield from self.__ls__gen__
        except BaseException as e:
            _cleanup_traceback(e)
            self._end_trace(error=e)
            raise
        else:
            self._end_trace()

    def __enter__(self):
        self.__ls_stream__.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            return self.__ls_stream__.__exit__(exc_type, exc_val, exc_tb)
        finally:
            self._end_trace(error=exc_val if exc_type else None)


class _TracedAsyncStream(_TracedStreamBase, Generic[T]):
    """A wrapper for asynchronous stream objects that handles tracing."""

    def __init__(
        self,
        stream: AsyncIterator[T],
        trace_container: _TraceableContainer,
        reduce_fn: Optional[Callable] = None,
        process_chunk: Optional[Callable] = None,
    ):
        super().__init__(
            stream=stream,
            trace_container=trace_container,
            reduce_fn=reduce_fn,
        )
        self.__ls_stream__ = stream
        self.__ls_gen = _process_async_iterator(
            generator=self.__ls_stream__,
            run_container=self.__ls_trace_container__,
            is_llm_run=self.__is_llm_run__,
            accepts_context=aitertools.asyncio_accepts_context(),
            results=self.__ls_accumulated_output__,
            process_chunk=process_chunk,
        )

    async def _aend_trace(self, error: Optional[BaseException] = None):
        ctx = copy_context()
        await asyncio.shield(aitertools.aio_to_thread(ctx, self._end_trace, error))
        _set_tracing_context(get_tracing_context(ctx))

    async def __anext__(self) -> T:
        try:
            return cast(T, await aitertools.py_anext(self.__ls_gen))
        except StopAsyncIteration:
            await self._aend_trace()
            raise

    async def __aiter__(self) -> AsyncIterator[T]:
        try:
            async for item in self.__ls_gen:
                yield item
        except BaseException:
            await self._aend_trace()
            raise
        else:
            await self._aend_trace()

    async def __aenter__(self):
        await self.__ls_stream__.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            return await self.__ls_stream__.__aexit__(exc_type, exc_val, exc_tb)
        finally:
            await self._aend_trace()


def _get_function_result(results: list, reduce_fn: Callable) -> Any:
    if results:
        if reduce_fn is not None:
            try:
                return reduce_fn(results)
            except BaseException as e:
                LOGGER.error(e)
                return results
        else:
            return results


def _cleanup_traceback(e: BaseException):
    tb_ = e.__traceback__
    if tb_:
        while tb_.tb_next is not None and tb_.tb_frame.f_code.co_filename.endswith(
            _EXCLUDED_FRAME_FNAME
        ):
            tb_ = tb_.tb_next
        e.__traceback__ = tb_


def _is_otel_available() -> bool:
    """Cache for otel import check."""
    global _OTEL_AVAILABLE
    if _OTEL_AVAILABLE is None:
        try:
            import opentelemetry.trace  # noqa: F401

            _OTEL_AVAILABLE = True
        except ImportError:
            _OTEL_AVAILABLE = False
    return _OTEL_AVAILABLE


def _maybe_create_otel_context(run_tree: Optional[run_trees.RunTree]):
    """Create OpenTelemetry context manager if OTEL is enabled and available.

    Args:
        run_tree: The current run tree.

    Returns:
        Context manager for use_span or None if not available.
    """
    if not run_tree or run_tree.client.otel_exporter is None:
        return None
    if not _is_otel_available():
        return None

    from opentelemetry.trace import (  # type: ignore[import]
        NonRecordingSpan,
        SpanContext,
        TraceFlags,
        TraceState,
        use_span,
    )

    from langsmith._internal._otel_utils import (
        get_otel_span_id_from_uuid,
        get_otel_trace_id_from_uuid,
    )

    trace_id = get_otel_trace_id_from_uuid(run_tree.trace_id)
    span_id = get_otel_span_id_from_uuid(run_tree.id)

    span_context = SpanContext(
        trace_id=trace_id,
        span_id=span_id,
        is_remote=False,
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        trace_state=TraceState(),
    )

    non_recording_span = NonRecordingSpan(span_context)
    return use_span(non_recording_span)


# For backwards compatibility
_PROJECT_NAME = _context._PROJECT_NAME
_TAGS = _context._TAGS
_METADATA = _context._METADATA
_TRACING_ENABLED = _context._TRACING_ENABLED
_CLIENT = _context._CLIENT
_PARENT_RUN_TREE_REF = _context._PARENT_RUN_TREE_REF



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from datetime import datetime
from typing_extensions import Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["DatasetRetrieveCsvParams"]


class DatasetRetrieveCsvParams(TypedDict, total=False):
    as_of: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]
    """Only modifications made on or before this time are included.

    If None, the latest version of the dataset is used.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable, Optional
from datetime import datetime
from typing_extensions import Literal, Annotated, TypedDict

from ..._utils import PropertyInfo

__all__ = ["InsightCreateParams"]


class InsightCreateParams(TypedDict, total=False):
    attribute_schemas: Optional[Dict[str, object]]

    cluster_model: Optional[str]

    config_id: Optional[str]

    end_time: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]

    filter: Optional[str]

    hierarchy: Optional[Iterable[int]]

    is_scheduled: bool

    last_n_hours: Optional[int]

    model: Literal["openai", "anthropic"]

    name: Optional[str]

    partitions: Optional[Dict[str, str]]

    sample: Optional[float]

    start_time: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]

    summary_model: Optional[str]

    summary_prompt: Optional[str]

    user_context: Optional[Dict[str, str]]

    validate_model_secrets: bool



"""OpenTelemetry integration for LangSmith."""

import logging
from typing import Optional, cast

from langsmith import utils as ls_utils

from .processor import OtelExporter, OtelSpanProcessor

logger = logging.getLogger(__name__)

__all__ = ["configure", "OtelSpanProcessor", "OtelExporter"]


def configure(
    api_key: Optional[str] = None,
    project_name: Optional[str] = None,
    SpanProcessor: Optional[type] = None,
) -> bool:
    """Configure OpenTelemetry with LangSmith as the `TracerProvider`.

    Initializes OpenTelemetry with LangSmith as the primary and only `TracerProvider`.

    Usage:
        >>> from langsmith.integrations.otel import configure
        >>> configure(  # doctest: +SKIP
        ...     api_key="your-api-key", project_name="your-project"
        ... )

        Using environment variables:
        >>> # Set LANGSMITH_API_KEY and LANGSMITH_PROJECT
        >>> configure()  # Will use env vars  # doctest: +SKIP

    !!! warning

        This function is only for when LangSmith is your ONLY OpenTelemetry source.

        It sets the global TracerProvider, which can only be done once per application.

    This function will fail if OpenTelemetry is already initialized with another
    `TracerProvider` (you cannot override an existing `TracerProvider`).

    If you already have OpenTelemetry set up with other tools, use `OtelSpanProcessor`
    directly to add LangSmith to your existing setup:

    !!! example "Adding LangSmith to existing OTEL setup"
        ```python
        from opentelemetry import trace
        from langsmith.integrations.otel.processor import OtelSpanProcessor

        # Use your existing provider (already initialized)
        provider = trace.get_tracer_provider()

        # Add LangSmith processor to existing provider
        langsmith_processor = OtelSpanProcessor(
            api_key="your-api-key", project="your-project"
        )
        provider.add_span_processor(langsmith_processor)
        ```

    Args:
        api_key: LangSmith API key. Defaults to `LANGSMITH_API_KEY` env var.
        project_name: Project name. Defaults to `LANGSMITH_PROJECT` env var.
        SpanProcessor: Span processor class to use. Defaults to `BatchSpanProcessor`.

    Returns:
        `True` if configuration succeeded, `False` if `TracerProvider` already exists.
    """
    try:
        from opentelemetry import trace
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.trace import NoOpTracer, ProxyTracer, ProxyTracerProvider

        existing_provider = cast(TracerProvider, trace.get_tracer_provider())
        tracer = existing_provider.get_tracer(__name__)

        # Check if OpenTelemetry is in its default uninitialized state
        # (ProxyTracerProvider with NoOpTracer means no real TracerProvider was set)
        if (
            isinstance(existing_provider, ProxyTracerProvider)
            and hasattr(tracer, "_tracer")
            and isinstance(
                cast(
                    ProxyTracer,  # type: ignore[attr-defined, name-defined]
                    tracer,
                )._tracer,
                NoOpTracer,
            )
        ):
            # Safe to set TracerProvider since none exists yet
            provider = TracerProvider()
            trace.set_tracer_provider(provider)
        else:
            logger.warning(
                "OpenTelemetry TracerProvider is already set. "
                "Cannot override existing TracerProvider. Use OtelSpanProcessor "
                "directly to add LangSmith to your existing provider instead."
            )
            return False

        api_key = api_key or ls_utils.get_api_key(None)
        if not api_key:
            return False

        project_name = project_name or ls_utils.get_tracer_project()

        processor = OtelSpanProcessor(
            api_key=api_key, project=project_name, SpanProcessor=SpanProcessor
        )
        provider.add_span_processor(processor)  # type: ignore
        return True
    except Exception as e:
        logger.warning("Failed to initialize Otel for LangSmith:", e)
        return False



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from datetime import datetime

import httpx

from ..._types import Body, Omit, Query, Headers, NotGiven, SequenceNotStr, omit, not_given
from ..._utils import path_template, maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ..._base_client import make_request_options
from ...types.datasets import split_create_params, split_retrieve_params
from ...types.datasets.split_create_response import SplitCreateResponse
from ...types.datasets.split_retrieve_response import SplitRetrieveResponse

__all__ = ["SplitsResource", "AsyncSplitsResource"]


class SplitsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> SplitsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/stainless-sdks/langchain-python#accessing-raw-response-data-eg-headers
        """
        return SplitsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> SplitsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/stainless-sdks/langchain-python#with_streaming_response
        """
        return SplitsResourceWithStreamingResponse(self)

    def create(
        self,
        dataset_id: str,
        *,
        examples: SequenceNotStr[str],
        split_name: str,
        remove: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SplitCreateResponse:
        """
        Update Dataset Splits

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not dataset_id:
            raise ValueError(f"Expected a non-empty value for `dataset_id` but received {dataset_id!r}")
        return self._put(
            path_template("/api/v1/datasets/{dataset_id}/splits", dataset_id=dataset_id),
            body=maybe_transform(
                {
                    "examples": examples,
                    "split_name": split_name,
                    "remove": remove,
                },
                split_create_params.SplitCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SplitCreateResponse,
        )

    def retrieve(
        self,
        dataset_id: str,
        *,
        as_of: Union[Union[str, datetime], str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SplitRetrieveResponse:
        """
        Get Dataset Splits

        Args:
          as_of: Only modifications made on or before this time are included. If None, the latest
              version of the dataset is used.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not dataset_id:
            raise ValueError(f"Expected a non-empty value for `dataset_id` but received {dataset_id!r}")
        return self._get(
            path_template("/api/v1/datasets/{dataset_id}/splits", dataset_id=dataset_id),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform({"as_of": as_of}, split_retrieve_params.SplitRetrieveParams),
            ),
            cast_to=SplitRetrieveResponse,
        )


class AsyncSplitsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncSplitsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/stainless-sdks/langchain-python#accessing-raw-response-data-eg-headers
        """
        return AsyncSplitsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncSplitsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/stainless-sdks/langchain-python#with_streaming_response
        """
        return AsyncSplitsResourceWithStreamingResponse(self)

    async def create(
        self,
        dataset_id: str,
        *,
        examples: SequenceNotStr[str],
        split_name: str,
        remove: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SplitCreateResponse:
        """
        Update Dataset Splits

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not dataset_id:
            raise ValueError(f"Expected a non-empty value for `dataset_id` but received {dataset_id!r}")
        return await self._put(
            path_template("/api/v1/datasets/{dataset_id}/splits", dataset_id=dataset_id),
            body=await async_maybe_transform(
                {
                    "examples": examples,
                    "split_name": split_name,
                    "remove": remove,
                },
                split_create_params.SplitCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SplitCreateResponse,
        )

    async def retrieve(
        self,
        dataset_id: str,
        *,
        as_of: Union[Union[str, datetime], str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SplitRetrieveResponse:
        """
        Get Dataset Splits

        Args:
          as_of: Only modifications made on or before this time are included. If None, the latest
              version of the dataset is used.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not dataset_id:
            raise ValueError(f"Expected a non-empty value for `dataset_id` but received {dataset_id!r}")
        return await self._get(
            path_template("/api/v1/datasets/{dataset_id}/splits", dataset_id=dataset_id),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=await async_maybe_transform({"as_of": as_of}, split_retrieve_params.SplitRetrieveParams),
            ),
            cast_to=SplitRetrieveResponse,
        )


class SplitsResourceWithRawResponse:
    def __init__(self, splits: SplitsResource) -> None:
        self._splits = splits

        self.create = to_raw_response_wrapper(
            splits.create,
        )
        self.retrieve = to_raw_response_wrapper(
            splits.retrieve,
        )


class AsyncSplitsResourceWithRawResponse:
    def __init__(self, splits: AsyncSplitsResource) -> None:
        self._splits = splits

        self.create = async_to_raw_response_wrapper(
            splits.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            splits.retrieve,
        )


class SplitsResourceWithStreamingResponse:
    def __init__(self, splits: SplitsResource) -> None:
        self._splits = splits

        self.create = to_streamed_response_wrapper(
            splits.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            splits.retrieve,
        )


class AsyncSplitsResourceWithStreamingResponse:
    def __init__(self, splits: AsyncSplitsResource) -> None:
        self._splits = splits

        self.create = async_to_streamed_response_wrapper(
            splits.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            splits.retrieve,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from datetime import datetime

from .._models import BaseModel

__all__ = ["InfoListResponse", "BatchIngestConfig", "CustomerInfo"]


class BatchIngestConfig(BaseModel):
    """Batch ingest config."""

    scale_down_nempty_trigger: Optional[int] = None

    scale_up_nthreads_limit: Optional[int] = None

    scale_up_qsize_trigger: Optional[int] = None

    size_limit: Optional[int] = None

    size_limit_bytes: Optional[int] = None

    use_multipart_endpoint: Optional[bool] = None


class CustomerInfo(BaseModel):
    """Customer info."""

    customer_id: str

    customer_name: str


class InfoListResponse(BaseModel):
    """The LangSmith server info."""

    version: str

    batch_ingest_config: Optional[BatchIngestConfig] = None
    """Batch ingest config."""

    customer_info: Optional[CustomerInfo] = None
    """Customer info."""

    git_sha: Optional[str] = None

    instance_flags: Optional[Dict[str, object]] = None

    license_expiration_time: Optional[datetime] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional
from typing_extensions import Literal, Required, TypedDict

from ..._types import SequenceNotStr

__all__ = ["GroupRunsParams"]


class GroupRunsParams(TypedDict, total=False):
    group_by: Required[Literal["run_metadata", "example_metadata"]]

    metadata_key: Required[str]

    session_ids: Required[SequenceNotStr[str]]

    filters: Optional[Dict[str, SequenceNotStr[str]]]

    limit: int

    offset: int

    per_group_limit: int

    preview: bool



"""OpenTelemetry span processor and exporter for LangSmith."""

import logging
import warnings
from typing import Any, Optional
from urllib.parse import urljoin

from langsmith import utils as ls_utils

try:
    from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
    from opentelemetry.sdk.trace.export import BatchSpanProcessor

    OTEL_AVAILABLE = True
except ImportError:
    warnings.warn(
        "OpenTelemetry packages are not installed. "
        "Install optional OpenTelemetry dependencies with: "
        "pip install langsmith[otel]",
        UserWarning,
        stacklevel=2,
    )

    class OTLPSpanExporter:  # type: ignore[no-redef]
        """Mock otlp span exporter class."""

        def __init__(self, *args, **kwargs):
            """Mock init method."""
            raise ImportError(
                "OpenTelemetry packages are not installed. "
                "Install optional OpenTelemetry dependencies with: "
                "pip install langsmith[otel]"
            )

    class BatchSpanProcessor:  # type: ignore[no-redef]
        """Mock batch span processor class."""

        def __init__(self, *args, **kwargs):
            """Mock init method."""
            raise ImportError(
                "OpenTelemetry packages are not installed. "
                "Install optional OpenTelemetry dependencies with: "
                "pip install langsmith[otel]"
            )

    class trace:
        """Mock trace class."""

        @staticmethod
        def get_tracer_provider():
            """Mock get tracer provider method."""
            raise ImportError(
                "OpenTelemetry packages are not installed. "
                "Install optional OpenTelemetry dependencies with: "
                "pip install langsmith[otel]"
            )

    OTEL_AVAILABLE = False

LANGSMITH_METADATA_PREFIX = "langsmith.metadata"


class OtelExporter(OTLPSpanExporter):
    """A subclass of `OTLPSpanExporter` configured for LangSmith.

    Environment Variables:

    - `LANGSMITH_API_KEY`: Your LangSmith API key.
    - `LANGSMITH_ENDPOINT`: Base URL for LangSmith API (defaults to `https://api.smith.langchain.com`).
    - `LANGSMITH_PROJECT`: Project identifier.
    """

    def __init__(
        self,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        project: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
        **kwargs,
    ):
        """Initialize the `OtelExporter`.

        Args:
            url: OTLP endpoint URL. Defaults to `{LANGSMITH_ENDPOINT}/otel/v1/traces`.
            api_key: LangSmith API key. Defaults to `LANGSMITH_API_KEY` env var.
            parent: Parent identifier (e.g., `'project_name:test'`).

                Defaults to `LANGSMITH_PARENT` env var.
            headers: Additional headers to include in requests.
            **kwargs: Additional arguments passed to `OTLPSpanExporter`.
        """
        base_url = ls_utils.get_api_url(None)
        # Ensure base_url ends with / for proper joining
        if not base_url.endswith("/"):
            base_url += "/"
        endpoint = url or urljoin(base_url, "otel/v1/traces")
        api_key = api_key or ls_utils.get_api_key(None)
        project = project or ls_utils.get_tracer_project()
        headers = headers or {}

        if not api_key:
            raise ValueError(
                "API key is required. Provide it via api_key parameter or "
                "LANGSMITH_API_KEY environment variable."
            )

        if not project:
            project = "default"
            logging.info(
                "No project specified, using default. "
                "Configure with LANGSMITH_PROJECT environment variable or "
                "project parameter."
            )

        exporter_headers = {
            "x-api-key": api_key,
            **headers,
        }

        if project:
            exporter_headers["Langsmith-Project"] = project

        self.project = project

        super().__init__(endpoint=endpoint, headers=exporter_headers, **kwargs)


class OtelSpanProcessor:
    """A span processor for adding LangSmith to OpenTelemetry setups.

    This class combines the `OtelExporter` and `BatchSpanProcessor`
    into a single processor that can be added to any `TracerProvider`.

    Use this when:

    1. You already have OpenTelemetry initialized with other tools
    2. You want to add LangSmith alongside existing OTEL exporters

    Examples:
        # Fresh OpenTelemetry setup (LangSmith only):
        from langsmith.integrations.otel import configure
        configure(api_key="your-key", project="your-project")

        # Add LangSmith to existing OpenTelemetry setup:
        from opentelemetry import trace
        from langsmith.integrations.otel.processor import OtelSpanProcessor

        # Get your existing TracerProvider (already set by other tools)
        provider = trace.get_tracer_provider()

        # Add LangSmith processor alongside existing processors
        langsmith_processor = OtelSpanProcessor(
            project="your-project",
        )
        provider.add_span_processor(langsmith_processor)
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        project: Optional[str] = None,
        url: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
        SpanProcessor: Optional[type] = None,
    ):
        """Initialize the `OtelSpanProcessor`.

        Args:
            api_key: LangSmith API key. Defaults to `LANGSMITH_API_KEY` env var.
            project: Project identifier. Defaults to `LANGSMITH_PROJECT` env var.
            url: Base URL for LangSmith API. Defaults to `LANGSMITH_ENDPOINT` env var
                or `https://api.smith.langchain.com`.
            headers: Additional headers to include in requests.
            SpanProcessor: Optional span processor class. Defaults to
                `BatchSpanProcessor`.
        """
        # Create the exporter
        # Convert url to the full endpoint URL that OtelExporter expects
        exporter_url = None
        if url:
            exporter_url = f"{url.rstrip('/')}/otel/v1/traces"

        self._exporter = OtelExporter(
            url=exporter_url, api_key=api_key, project=project, headers=headers
        )

        # Create the processor chain
        if not OTEL_AVAILABLE:
            raise ImportError(
                "OpenTelemetry packages are not installed. "
                "Install optional OpenTelemetry dependencies with: "
                "pip install langsmith[otel]"
            )

        if SpanProcessor is None:
            SpanProcessor = BatchSpanProcessor

        self._processor = SpanProcessor(self._exporter)
        self._metadata: dict[str, Any] = {}

    def set_metadata(self, metadata: dict[str, Any]) -> None:
        """Set metadata attributes to propagate to all spans.

        Use this to ensure metadata like ``thread_id`` appears on every span
        in a trace, not just the root span.  This is required for LangSmith
        features (e.g. the threads view) that query runs by metadata fields.
        """
        self._metadata = metadata.copy()

    def on_start(self, span, parent_context=None):
        """Forward span start events to the inner processor."""
        if self._metadata:
            for key, value in self._metadata.items():
                if value is not None:
                    span.set_attribute(f"{LANGSMITH_METADATA_PREFIX}.{key}", value)
        self._processor.on_start(span, parent_context)

    def on_end(self, span):
        """Forward span end events to the inner processor."""
        self._processor.on_end(span)

    def _on_ending(self, span):
        """Forward span ending events to the inner processor."""
        if hasattr(self._processor, "_on_ending"):
            self._processor._on_ending(span)

    def shutdown(self):
        """Shutdown processor."""
        self._processor.shutdown()

    def force_flush(self, timeout_millis=30000):
        """Force flush the inner processor."""
        return self._processor.force_flush(timeout_millis)

    @property
    def exporter(self):
        """The underlying OtelExporter."""
        return self._exporter

    @property
    def processor(self):
        """The underlying span processor."""
        return self._processor



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Union, Optional
from datetime import datetime
from typing_extensions import Literal, TypeAlias

from .._models import BaseModel

__all__ = [
    "CustomChartsSection",
    "Chart",
    "ChartData",
    "ChartSeries",
    "ChartSeriesFilterDefinition",
    "ChartSeriesFilterDefinitionCustomChartFilterByTracingProject",
    "ChartSeriesFilterDefinitionCustomChartFilterByDataset",
    "ChartSeriesFilters",
    "ChartSeriesGroupBy",
    "ChartSeriesGroupByDefinition",
    "ChartSeriesGroupByDefinitionCustomChartGroupByPlain",
    "ChartSeriesGroupByDefinitionCustomChartGroupByComplex",
    "ChartSeriesMetricDefinition",
    "ChartSeriesMetricDefinitionCustomChartMetricCount",
    "ChartSeriesMetricDefinitionCustomChartMetricScalar",
    "ChartSeriesMetricDefinitionCustomChartMetricPercentile",
    "ChartSeriesMetricDefinitionCustomChartMetricPercentileParams",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutput",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominator",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricCount",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricScalar",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentile",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentileParams",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumerator",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricCount",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricScalar",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentile",
    "ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentileParams",
    "ChartCommonFilters",
    "SubSection",
    "SubSectionChart",
    "SubSectionChartData",
    "SubSectionChartSeries",
    "SubSectionChartSeriesFilterDefinition",
    "SubSectionChartSeriesFilterDefinitionCustomChartFilterByTracingProject",
    "SubSectionChartSeriesFilterDefinitionCustomChartFilterByDataset",
    "SubSectionChartSeriesFilters",
    "SubSectionChartSeriesGroupBy",
    "SubSectionChartSeriesGroupByDefinition",
    "SubSectionChartSeriesGroupByDefinitionCustomChartGroupByPlain",
    "SubSectionChartSeriesGroupByDefinitionCustomChartGroupByComplex",
    "SubSectionChartSeriesMetricDefinition",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricCount",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricScalar",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricPercentile",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricPercentileParams",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutput",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominator",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricCount",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricScalar",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentile",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentileParams",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumerator",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricCount",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricScalar",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentile",
    "SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentileParams",
    "SubSectionChartCommonFilters",
]


class ChartData(BaseModel):
    series_id: str

    timestamp: datetime

    value: Union[float, Dict[str, object], None] = None

    group: Optional[str] = None


class ChartSeriesFilterDefinitionCustomChartFilterByTracingProject(BaseModel):
    project_ids: List[str]

    source_type: Literal["tracing_project"]

    run_filter: Optional[str] = None

    trace_filter: Optional[str] = None

    tree_filter: Optional[str] = None


class ChartSeriesFilterDefinitionCustomChartFilterByDataset(BaseModel):
    dataset_ids: List[str]

    source_type: Literal["dataset"]


ChartSeriesFilterDefinition: TypeAlias = Union[
    ChartSeriesFilterDefinitionCustomChartFilterByTracingProject,
    ChartSeriesFilterDefinitionCustomChartFilterByDataset,
    None,
]


class ChartSeriesFilters(BaseModel):
    filter: Optional[str] = None

    session: Optional[List[str]] = None

    trace_filter: Optional[str] = None

    tree_filter: Optional[str] = None


class ChartSeriesGroupBy(BaseModel):
    """Include additional information about where the group_by param was set."""

    attribute: Literal["name", "run_type", "tag", "metadata"]

    max_groups: Optional[int] = None

    path: Optional[str] = None

    set_by: Optional[Literal["section", "series"]] = None


class ChartSeriesGroupByDefinitionCustomChartGroupByPlain(BaseModel):
    attribute: Literal["name", "run_type", "tag", "project", "status"]


class ChartSeriesGroupByDefinitionCustomChartGroupByComplex(BaseModel):
    attribute: Literal["metadata", "feedback_label"]

    path: str


ChartSeriesGroupByDefinition: TypeAlias = Union[
    ChartSeriesGroupByDefinitionCustomChartGroupByPlain, ChartSeriesGroupByDefinitionCustomChartGroupByComplex
]


class ChartSeriesMetricDefinitionCustomChartMetricCount(BaseModel):
    filter: Optional[str] = None

    type: Optional[Literal["count"]] = None


class ChartSeriesMetricDefinitionCustomChartMetricScalar(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    type: Literal["sum", "max", "min", "avg"]

    filter: Optional[str] = None


class ChartSeriesMetricDefinitionCustomChartMetricPercentileParams(BaseModel):
    p: int


class ChartSeriesMetricDefinitionCustomChartMetricPercentile(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    params: ChartSeriesMetricDefinitionCustomChartMetricPercentileParams

    type: Literal["percentile"]

    filter: Optional[str] = None


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricCount(BaseModel):
    filter: Optional[str] = None

    type: Optional[Literal["count"]] = None


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricScalar(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    type: Literal["sum", "max", "min", "avg"]

    filter: Optional[str] = None


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentileParams(BaseModel):
    p: int


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentile(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    params: ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentileParams

    type: Literal["percentile"]

    filter: Optional[str] = None


ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominator: TypeAlias = Union[
    ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricCount,
    ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricScalar,
    ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentile,
]


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricCount(BaseModel):
    filter: Optional[str] = None

    type: Optional[Literal["count"]] = None


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricScalar(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    type: Literal["sum", "max", "min", "avg"]

    filter: Optional[str] = None


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentileParams(BaseModel):
    p: int


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentile(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    params: ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentileParams

    type: Literal["percentile"]

    filter: Optional[str] = None


ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumerator: TypeAlias = Union[
    ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricCount,
    ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricScalar,
    ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentile,
]


class ChartSeriesMetricDefinitionCustomChartMetricRatioOutput(BaseModel):
    denominator: ChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominator

    numerator: ChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumerator

    type: Literal["ratio"]


ChartSeriesMetricDefinition: TypeAlias = Union[
    ChartSeriesMetricDefinitionCustomChartMetricCount,
    ChartSeriesMetricDefinitionCustomChartMetricScalar,
    ChartSeriesMetricDefinitionCustomChartMetricPercentile,
    ChartSeriesMetricDefinitionCustomChartMetricRatioOutput,
    None,
]


class ChartSeries(BaseModel):
    id: str

    name: str

    feedback_key: Optional[str] = None

    filter_definition: Optional[ChartSeriesFilterDefinition] = None

    filters: Optional[ChartSeriesFilters] = None

    group_by: Optional[ChartSeriesGroupBy] = None
    """Include additional information about where the group_by param was set."""

    group_by_definitions: Optional[List[ChartSeriesGroupByDefinition]] = None

    metric: Optional[
        Literal[
            "run_count",
            "latency_p50",
            "latency_p99",
            "latency_avg",
            "first_token_p50",
            "first_token_p99",
            "total_tokens",
            "prompt_tokens",
            "completion_tokens",
            "median_tokens",
            "completion_tokens_p50",
            "prompt_tokens_p50",
            "tokens_p99",
            "completion_tokens_p99",
            "prompt_tokens_p99",
            "feedback",
            "feedback_score_avg",
            "feedback_values",
            "total_cost",
            "prompt_cost",
            "completion_cost",
            "error_rate",
            "streaming_rate",
            "cost_p50",
            "cost_p99",
        ]
    ] = None
    """Metrics you can chart.

    Feedback metrics are not available for organization-scoped charts.
    """

    metric_definition: Optional[ChartSeriesMetricDefinition] = None

    project_metric: Optional[
        Literal[
            "memory_usage",
            "cpu_usage",
            "disk_usage",
            "restart_count",
            "replica_count",
            "worker_count",
            "lg_run_count",
            "responses_per_second",
            "error_responses_per_second",
            "p95_latency",
        ]
    ] = None
    """LGP Metrics you can chart."""

    workspace_id: Optional[str] = None


class ChartCommonFilters(BaseModel):
    filter: Optional[str] = None

    session: Optional[List[str]] = None

    trace_filter: Optional[str] = None

    tree_filter: Optional[str] = None


class Chart(BaseModel):
    id: str

    chart_type: Literal["line", "bar", "table", "kpi", "top-k", "pie"]
    """Enum for custom chart types."""

    data: List[ChartData]

    index: int

    series: List[ChartSeries]

    title: str

    common_filters: Optional[ChartCommonFilters] = None

    description: Optional[str] = None

    metadata: Optional[Dict[str, object]] = None


class SubSectionChartData(BaseModel):
    series_id: str

    timestamp: datetime

    value: Union[float, Dict[str, object], None] = None

    group: Optional[str] = None


class SubSectionChartSeriesFilterDefinitionCustomChartFilterByTracingProject(BaseModel):
    project_ids: List[str]

    source_type: Literal["tracing_project"]

    run_filter: Optional[str] = None

    trace_filter: Optional[str] = None

    tree_filter: Optional[str] = None


class SubSectionChartSeriesFilterDefinitionCustomChartFilterByDataset(BaseModel):
    dataset_ids: List[str]

    source_type: Literal["dataset"]


SubSectionChartSeriesFilterDefinition: TypeAlias = Union[
    SubSectionChartSeriesFilterDefinitionCustomChartFilterByTracingProject,
    SubSectionChartSeriesFilterDefinitionCustomChartFilterByDataset,
    None,
]


class SubSectionChartSeriesFilters(BaseModel):
    filter: Optional[str] = None

    session: Optional[List[str]] = None

    trace_filter: Optional[str] = None

    tree_filter: Optional[str] = None


class SubSectionChartSeriesGroupBy(BaseModel):
    """Include additional information about where the group_by param was set."""

    attribute: Literal["name", "run_type", "tag", "metadata"]

    max_groups: Optional[int] = None

    path: Optional[str] = None

    set_by: Optional[Literal["section", "series"]] = None


class SubSectionChartSeriesGroupByDefinitionCustomChartGroupByPlain(BaseModel):
    attribute: Literal["name", "run_type", "tag", "project", "status"]


class SubSectionChartSeriesGroupByDefinitionCustomChartGroupByComplex(BaseModel):
    attribute: Literal["metadata", "feedback_label"]

    path: str


SubSectionChartSeriesGroupByDefinition: TypeAlias = Union[
    SubSectionChartSeriesGroupByDefinitionCustomChartGroupByPlain,
    SubSectionChartSeriesGroupByDefinitionCustomChartGroupByComplex,
]


class SubSectionChartSeriesMetricDefinitionCustomChartMetricCount(BaseModel):
    filter: Optional[str] = None

    type: Optional[Literal["count"]] = None


class SubSectionChartSeriesMetricDefinitionCustomChartMetricScalar(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    type: Literal["sum", "max", "min", "avg"]

    filter: Optional[str] = None


class SubSectionChartSeriesMetricDefinitionCustomChartMetricPercentileParams(BaseModel):
    p: int


class SubSectionChartSeriesMetricDefinitionCustomChartMetricPercentile(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    params: SubSectionChartSeriesMetricDefinitionCustomChartMetricPercentileParams

    type: Literal["percentile"]

    filter: Optional[str] = None


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricCount(BaseModel):
    filter: Optional[str] = None

    type: Optional[Literal["count"]] = None


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricScalar(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    type: Literal["sum", "max", "min", "avg"]

    filter: Optional[str] = None


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentileParams(
    BaseModel
):
    p: int


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentile(
    BaseModel
):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    params: (
        SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentileParams
    )

    type: Literal["percentile"]

    filter: Optional[str] = None


SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominator: TypeAlias = Union[
    SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricCount,
    SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricScalar,
    SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominatorCustomChartMetricPercentile,
]


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricCount(BaseModel):
    filter: Optional[str] = None

    type: Optional[Literal["count"]] = None


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricScalar(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    type: Literal["sum", "max", "min", "avg"]

    filter: Optional[str] = None


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentileParams(
    BaseModel
):
    p: int


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentile(BaseModel):
    field: Literal[
        "latency_seconds",
        "first_token_seconds",
        "total_tokens",
        "prompt_tokens",
        "completion_tokens",
        "total_cost",
        "prompt_cost",
        "completion_cost",
    ]

    params: SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentileParams

    type: Literal["percentile"]

    filter: Optional[str] = None


SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumerator: TypeAlias = Union[
    SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricCount,
    SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricScalar,
    SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumeratorCustomChartMetricPercentile,
]


class SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutput(BaseModel):
    denominator: SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputDenominator

    numerator: SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutputNumerator

    type: Literal["ratio"]


SubSectionChartSeriesMetricDefinition: TypeAlias = Union[
    SubSectionChartSeriesMetricDefinitionCustomChartMetricCount,
    SubSectionChartSeriesMetricDefinitionCustomChartMetricScalar,
    SubSectionChartSeriesMetricDefinitionCustomChartMetricPercentile,
    SubSectionChartSeriesMetricDefinitionCustomChartMetricRatioOutput,
    None,
]


class SubSectionChartSeries(BaseModel):
    id: str

    name: str

    feedback_key: Optional[str] = None

    filter_definition: Optional[SubSectionChartSeriesFilterDefinition] = None

    filters: Optional[SubSectionChartSeriesFilters] = None

    group_by: Optional[SubSectionChartSeriesGroupBy] = None
    """Include additional information about where the group_by param was set."""

    group_by_definitions: Optional[List[SubSectionChartSeriesGroupByDefinition]] = None

    metric: Optional[
        Literal[
            "run_count",
            "latency_p50",
            "latency_p99",
            "latency_avg",
            "first_token_p50",
            "first_token_p99",
            "total_tokens",
            "prompt_tokens",
            "completion_tokens",
            "median_tokens",
            "completion_tokens_p50",
            "prompt_tokens_p50",
            "tokens_p99",
            "completion_tokens_p99",
            "prompt_tokens_p99",
            "feedback",
            "feedback_score_avg",
            "feedback_values",
            "total_cost",
            "prompt_cost",
            "completion_cost",
            "error_rate",
            "streaming_rate",
            "cost_p50",
            "cost_p99",
        ]
    ] = None
    """Metrics you can chart.

    Feedback metrics are not available for organization-scoped charts.
    """

    metric_definition: Optional[SubSectionChartSeriesMetricDefinition] = None

    project_metric: Optional[
        Literal[
            "memory_usage",
            "cpu_usage",
            "disk_usage",
            "restart_count",
            "replica_count",
            "worker_count",
            "lg_run_count",
            "responses_per_second",
            "error_responses_per_second",
            "p95_latency",
        ]
    ] = None
    """LGP Metrics you can chart."""

    workspace_id: Optional[str] = None


class SubSectionChartCommonFilters(BaseModel):
    filter: Optional[str] = None

    session: Optional[List[str]] = None

    trace_filter: Optional[str] = None

    tree_filter: Optional[str] = None


class SubSectionChart(BaseModel):
    id: str

    chart_type: Literal["line", "bar", "table", "kpi", "top-k", "pie"]
    """Enum for custom chart types."""

    data: List[SubSectionChartData]

    index: int

    series: List[SubSectionChartSeries]

    title: str

    common_filters: Optional[SubSectionChartCommonFilters] = None

    description: Optional[str] = None

    metadata: Optional[Dict[str, object]] = None


class SubSection(BaseModel):
    id: str

    charts: List[SubSectionChart]

    index: int

    title: str

    description: Optional[str] = None


class CustomChartsSection(BaseModel):
    id: str

    charts: List[Chart]

    title: str

    description: Optional[str] = None

    index: Optional[int] = None

    session_id: Optional[str] = None

    sub_sections: Optional[List[SubSection]] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from datetime import datetime

from ..._models import BaseModel

__all__ = ["ComparativeCreateResponse"]


class ComparativeCreateResponse(BaseModel):
    """ComparativeExperiment schema."""

    id: str

    created_at: datetime

    modified_at: datetime

    reference_dataset_id: str

    tenant_id: str

    description: Optional[str] = None

    extra: Optional[Dict[str, object]] = None

    name: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["SessionSortableColumns"]

SessionSortableColumns: TypeAlias = Literal[
    "name", "start_time", "last_run_start_time", "latency_p50", "latency_p99", "error_rate", "feedback", "runs_count"
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Optional
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from .._types import SequenceNotStr
from .._utils import PropertyInfo

__all__ = ["DatasetCloneParams"]


class DatasetCloneParams(TypedDict, total=False):
    source_dataset_id: Required[str]

    target_dataset_id: Required[str]

    as_of: Annotated[Union[Union[str, datetime], str, None], PropertyInfo(format="iso8601")]
    """Only modifications made on or before this time are included.

    If None, the latest version of the dataset is used.
    """

    examples: SequenceNotStr[str]

    split: Union[str, SequenceNotStr[str], None]

    tag_value_ids: Optional[SequenceNotStr[str]]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from ..._models import BaseModel

__all__ = ["InsightDeleteResponse"]


class InsightDeleteResponse(BaseModel):
    """Response to delete a session cluster job."""

    id: str

    message: str



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal, Required, TypedDict

__all__ = ["RunStatsGroupByParam"]


class RunStatsGroupByParam(TypedDict, total=False):
    """Group by param for run stats."""

    attribute: Required[Literal["name", "run_type", "tag", "metadata"]]

    max_groups: int

    path: Optional[str]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Required, TypedDict

from .._types import FileTypes, SequenceNotStr
from .data_type import DataType

__all__ = ["DatasetUploadParams"]


class DatasetUploadParams(TypedDict, total=False):
    file: Required[FileTypes]

    input_keys: Required[SequenceNotStr[str]]

    data_type: DataType
    """Enum for dataset data types."""

    description: Optional[str]

    input_key_mappings: Optional[str]

    inputs_schema_definition: Optional[str]

    metadata_key_mappings: Optional[str]

    metadata_keys: SequenceNotStr[str]

    name: Optional[str]

    output_key_mappings: Optional[str]

    output_keys: SequenceNotStr[str]

    outputs_schema_definition: Optional[str]

    tag_value_ids: Optional[str]

    transformations: Optional[str]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["SortByDatasetColumn"]

SortByDatasetColumn: TypeAlias = Literal[
    "name", "created_at", "last_session_start_time", "example_count", "session_count", "modified_at"
]
